"""Implementation of Results Base Class.

Peter Schubert, HHU Duesseldorf, CCB, Mai 2024
"""

import re
import numpy as np
import pandas as pd
import scipy
import json
import matplotlib as mpl
import matplotlib.pyplot as plt

from abc import ABC, abstractmethod

import f2xba.prefixes as pf


class Results(ABC):
    """Optimization results processing for extended models generated by f2xba.

    The Results class is parent to the `FbaResults`, `EcmResults`
    and `RbaResults` classes. It provides common functionality for processing
    optimization results, produced by cobrapy or gurobipy, during optimization of
    extended models for a single or several (media) conditions.
    """


    def __init__(self, optim, results, df_mpmf=None):
        """Instantiate the Results base class.

        This instantiation is invoked by child classes and requires a
        reference to the Optimize instance, for retrieving model specific
        data. The results is a dictionary containing one or several
        optimization results. Optionally, measured protein levels can
        be provided for one or several conditions in mg measured protein
        to gram total protein. Individual proteins are identified by gene locus.
        The proteomics table may include optional information columns named
        'uniprot', 'description', 'gene_name', 'mw', 'avg_mpmf', 'rank'.

        :param optim: xxxOptimization instance
        :type optim: :class:`Optimize`
        :param results: conditions with optimization results
        :type results: dict (key: str, val: :class:`Solution`)
        :param df_mpmf: measured protein mass fractions for conditions (default: None)
        :type df_mpmf: pandas.DataFrame
        """
        self.optim = optim
        self.results = results
        if type(df_mpmf) is pd.DataFrame:
            self.df_mpmf = df_mpmf
            self.gene2uid = df_mpmf['uniprot'] if 'uniprot' in df_mpmf.columns else {}
            self.gene2name = df_mpmf['gene_name'] if 'gene_name' in df_mpmf.columns else {}
            self.gene2descr = df_mpmf['description'] if 'description' in df_mpmf.columns else {}
            self.gene2mw_kda = df_mpmf['mw_Da'] / 1000.0 if 'mw_Da' in df_mpmf.columns else {}
            self.gene2mpmf = df_mpmf['avg_mpmf'] if 'avg_mpmf' in df_mpmf.columns else {}
            self.gene2rank = df_mpmf['rank'] if 'rank' in df_mpmf.columns else {}

    @abstractmethod
    def get_fluxes(self, solution):
        pass

    @abstractmethod
    def get_net_fluxes(self, solution):
        pass

    def collect_fluxes(self, net=False):
        """Extract predicted reaction fluxes from optimization results.

        In extended models, reactions may be divided into forward and reverse
        directions, as well as into reactions catalyzed by isoenzymes. When the "net"
        parameter is set to "true," the fluxes of split reactions are combined, a
        result that is comparable to flux results from FBA optimizations.
        Information columns are incorporated to provide context and facilitate data filtration.

        :param bool net: summarize fluxes on reaction level (default: False)
        :return: table with predicted reaction fluxes
        :rtype: pandas.DataFrame
        """
        if net is True:
            info_cols = ['reaction_str', 'gpr']
        else:
            info_cols = ['reaction_str', 'net_rid', 'gpr']
        n_info_cols = len(info_cols)
        df_fluxes = None
        for condition, solution in self.results.items():
            df = self.get_net_fluxes(solution) if net is True else self.get_fluxes(solution)
            if df_fluxes is None:
                df_fluxes = df[info_cols + ['mmol_per_gDWh']].copy()
            else:
                df_fluxes = pd.concat([df_fluxes, df['mmol_per_gDWh']], axis=1)
            df_fluxes.rename(columns={'mmol_per_gDWh': f'{condition}'}, inplace=True)
        mean = df_fluxes.iloc[:, n_info_cols:].mean(axis=1).values
        stdev = df_fluxes.iloc[:, n_info_cols:].std(axis=1).values
        df_fluxes.insert(n_info_cols, 'mean mmol_per_gDWh', mean)
        df_fluxes.insert(n_info_cols + 1, 'abs_mean mmol_per_gDWh', abs(mean))
        df_fluxes.insert(n_info_cols + 2, 'stdev', stdev)
        df_fluxes.sort_values(by='abs_mean mmol_per_gDWh', ascending=False, inplace=True)
        rank = np.array(range(1, len(df_fluxes) + 1))
        df_fluxes.insert(n_info_cols, column='rank', value=rank)
        df_fluxes.index.name = 'rid'
        return df_fluxes

    @abstractmethod
    def collect_protein_results(self):
        """Extract predicted protein concentrations from optimization results.

        Information columns are incorporated to provide context and facilitate data filtration.

        :return: table with predicted protein concentrations
        :rtype: pandas.DataFrame
        """
        return pd.DataFrame()

    @abstractmethod
    def get_predicted_protein_data(self, solution):
        return pd.DataFrame()

    def _collect_protein_results(self, units):
        """Collect protein results, encriched with data from proteomics data set, if provided.
        """
        df_proteins = None
        n_info_cols = 0
        for condition, solution in self.results.items():
            df = self.get_predicted_protein_data(solution)
            if df_proteins is None:
                if hasattr(self, 'df_mpmf'):
                    data = []
                    for gene, row in df.iterrows():
                        uid = row['uniprot'] if type(row['uniprot']) is str else self.gene2uid.get(gene)
                        gene_name = row['gene_name'] if type(row['gene_name']) is str else self.gene2name.get(gene)
                        description = self.gene2descr.get(gene)
                        mw_kda = row['mw_kDa'] if 'mw_kDa' in row else self.gene2mw_kda.get(gene)
                        exp_rank = self.gene2rank.get(gene)
                        exp_mpmf = self.gene2mpmf.get(gene)
                        data.append([gene, uid, gene_name, description, mw_kda, exp_mpmf, exp_rank])
                    cols = ['gene', 'uniprot', 'gene_name', 'description', 'mw_kDa', 'exp_avg_mpmf', 'exp_rank']
                    df_proteins = pd.DataFrame(data, columns=cols).set_index('gene')
                    n_info_cols = len(cols) - 1
                else:
                    cols = [col for col in df.columns if col in {'uniprot', 'gene_name', 'mw_kDa'}]
                    df_proteins = df[cols].copy()
                    n_info_cols = len(cols)
            df_proteins = pd.concat([df_proteins, df[units]], axis=1)
            df_proteins.rename(columns={units: f'{condition}'}, inplace=True)

        mean = df_proteins.iloc[:, n_info_cols:].mean(axis=1).values
        stdev = df_proteins.iloc[:, n_info_cols:].std(axis=1).values
        df_proteins.insert(n_info_cols, f'mean {units}', mean)
        df_proteins.insert(n_info_cols + 1, 'stdev', stdev)
        df_proteins.index.name = 'gene'
        df_proteins.sort_values(by=f'mean {units}', ascending=False, inplace=True)
        rank = np.array(range(1, len(df_proteins) + 1))
        df_proteins.insert(n_info_cols, column='pred_rank', value=rank)
        return df_proteins

    def get_predicted_species_conc(self, solution):
        """Collect species concentrations (mmol/l) for a TD constraint model.

        :param solution: optimization solution
        :type solution: :class:`Solution`
        :return: table with predicted metabolite concentrations (mmol/l)
        :rtype: pandas.DataFrame
        """
        species_conc = {}
        for vid, val in solution.fluxes.items():
            if re.match(pf.V_LC_, vid):
                mid = re.sub(f'^{pf.V_LC_}', '', vid)
                name = self.optim.mid2name[mid]
                # name = self.optim.model.metabolites.get_by_id(mid).name
                species_conc[mid] = [name, np.exp(val) * 1e3]
        cols = ['name', 'mmol_per_l']
        df_conc = pd.DataFrame(species_conc.values(), index=list(species_conc), columns=cols)
        df_conc.index.name = 'mid'
        return df_conc

    def collect_species_conc(self):
        """Extract predicted species concentrations from optimization results.

        Information columns are incorporated to provide context and facilitate data filtration.

        :return: table with predicted species concentrations
        :rtype: pandas.DataFrame
        """
        df_conc = None
        n_info_cols = 0
        for condition, solution in self.results.items():
            df = self.get_predicted_species_conc(solution)
            if df_conc is None:
                df_conc = df.copy()
                n_info_cols = df_conc.shape[1] - 1
            else:
                df_conc = pd.concat([df_conc, df['mmol_per_l']], axis=1)
            df_conc.rename(columns={'mmol_per_l': f'{condition}'}, inplace=True)
        mean = df_conc.iloc[:, n_info_cols:].mean(axis=1).values
        stdev = df_conc.iloc[:, n_info_cols:].std(axis=1).values
        df_conc.insert(n_info_cols, 'mean mmol_per_l', mean)
        df_conc.insert(n_info_cols + 1, 'stdev', stdev)
        df_conc.sort_values(by='mean mmol_per_l', ascending=False, inplace=True)
        rank = np.array(range(1, len(df_conc) + 1))
        df_conc.insert(n_info_cols, 'rank', rank)
        df_conc.index.name = 'sid'
        return df_conc

    def get_rtype_condition_mpmf(self, r_type, condition):
        """For specified condition and reaction type extract protein mass fractions.

        predicted vs. experimental protein mass fractions used for correlation studies
        only return values for which experimental data exists

        :param str r_type: reaction type 'transport', 'metabolic' or 'process'
        :param str condition: (media) condition in optimization results
        :return: protein mass fractions (measured and predicted)
        :rtype: dict (key: gene/str, value: list(measured mpmf, predicted mpmf)/float/float
        """

        all_genes = set(self.optim.m_dict['fbcGeneProducts']['label'].values)
        tx_genes, metab_genes = self.optim.get_tx_metab_genes()
        pm_genes = all_genes.difference(tx_genes.union(metab_genes))

        mpmfs = {}
        if condition in self.results and condition in self.df_mpmf.columns:

            # determine gene set to be collected
            genes = set()
            if r_type == 'transport':
                genes = tx_genes
            elif r_type == 'metabolic':
                genes = metab_genes
            elif r_type == 'process':
                genes = pm_genes

            exp_mpmfs = self.df_mpmf[condition].to_dict()
            pred_mpmfs = self.get_predicted_protein_data(self.results[condition])['mg_per_gP'].to_dict()
            for gene, pred_mpmf in pred_mpmfs.items():
                if gene in exp_mpmfs and gene in genes:
                    mpmfs[gene] = [exp_mpmfs[gene], pred_mpmf]
        return mpmfs

    def report_gr_correlation(self, exp_grs):
        """Report on predicted to measured growth rates.

        :param exp_grs: (media) conditions with measured growth rates (h-1)
        :type exp_grs: dict (key: condition/str, val: growth rate)
        """
        conds = set(exp_grs).intersection(self.results)
        cond_exp_grs = np.array([exp_grs[cond] for cond in conds])
        cond_pred_grs = np.array([self.results[cond].objective_value for cond in conds])
        rel_error = abs(cond_exp_grs - cond_pred_grs) / cond_exp_grs
        r_value, p_value = scipy.stats.pearsonr(cond_exp_grs, cond_pred_grs)
        print(f'predicted grs ({len(cond_pred_grs)}) vs. experiment: r2 = {r_value ** 2:.4f}, p = {p_value:.2e}, '
              f'relative error = {np.mean(rel_error) * 100:.2f}%')

    def report_proteomics_correlation(self, scale='lin'):
        """Report on predicted to measured protein mass fractions.

        :param str scale: numerial value scale: 'lin' or 'log' (default: 'lin')
        """
        if hasattr(self, 'df_mpmf') is False:
            print('no proteomics data supplied for comparisson')
            return

        df_proteins = self.collect_protein_results()
        for condition in self.results:
            if condition in self.df_mpmf:
                exp_mpmfs = self.df_mpmf[condition].to_dict()
                pred_mpmfs = df_proteins[condition].to_dict()

                genes = set(exp_mpmfs.keys()).intersection(set(pred_mpmfs.keys()))
                xy = np.array([[exp_mpmfs[gene], pred_mpmfs[gene]] for gene in genes])
                x, y = (xy[:, 0], xy[:, 1]) if scale == 'lin' else self.get_log10_xy(xy)
                r_value, p_value = scipy.stats.pearsonr(x, y)
                print(f'{condition:25s}: r\N{SUPERSCRIPT TWO} = {r_value ** 2:.4f}, p = {p_value:.2e} '
                      f'({len(x):4d} proteins {scale} scale)')

    def report_protein_levels(self, condition):
        """Report on predicted and measured protein levels split per reaction type for given conditon.

        :param str condition: (media) condition in optimization results
        """
        # For measured proteins collect predicted mass fraction per type
        if hasattr(self, 'df_mpmf') is False:
            print('no proteomics data supplied for comparisson')
            return

        metab_gene2mpmfs = self.get_rtype_condition_mpmf('metabolic', condition)
        tx_gene2mpmfs = self.get_rtype_condition_mpmf('transport', condition)
        pm_gene2mpmfs = self.get_rtype_condition_mpmf('process', condition)
        all_gene2mpmfs = metab_gene2mpmfs | tx_gene2mpmfs | pm_gene2mpmfs
        metab_mpmfs = np.array(list(metab_gene2mpmfs.values()))
        tx_mpmfs = np.array(list(tx_gene2mpmfs.values()))
        pm_mpmfs = np.array(list(pm_gene2mpmfs.values()))
        all_mpmfs = np.array(list(all_gene2mpmfs.values()))

        # get total predicted protein mass, including proteins not measured
        all_pred = self.get_predicted_protein_data(self.results[condition])['mg_per_gP'].to_dict()
        dummy_pred = {gp: mpmf for gp, mpmf in all_pred.items() if 'dummy' in gp}
        other_pred_only = {gp: mpmf for gp, mpmf in all_pred.items() if
                           gp not in all_gene2mpmfs and gp not in dummy_pred}

        print(f'condition: {condition}')
        print(f'{len(all_pred):4d} proteins in model with total predicted mass fraction of '
              f'{sum(all_pred.values()):.1f} mg/gP')
        print(f'{"":5s}{len(all_mpmfs):4d} have been measured with mpmf of  {sum(all_mpmfs[:, 0]):5.1f} mg/gP vs. '
              f'{sum(all_mpmfs[:, 1]):.1f} mg/gP predicted')
        for p_class, mpmfs in {'metabolic': metab_mpmfs, 'transport': tx_mpmfs, 'processes': pm_mpmfs}.items():
            if len(mpmfs) > 0:
                print(f'{"":10s}{len(mpmfs):4d} {p_class} proteins measured {sum(mpmfs[:, 0]):5.1f} mg/gP '
                      f'vs. {sum(mpmfs[:, 1]):5.1f} mg/gP predicted')
        print(f'{"":5s}{len(dummy_pred) + len(other_pred_only):4d} proteins not measured vs. '
              f'{sum(dummy_pred.values()) + sum(other_pred_only.values()):5.1f} mg/gP predicted')
        for p_class, mpmfs in {'dummy ': dummy_pred, 'actual ': other_pred_only}.items():
            if len(mpmfs) > 0:
                print(f'{"":10s}{len(mpmfs):4d} {p_class} proteins {sum(mpmfs.values()):9.1f} mg/gP predicted')

        for scale in ['lin', 'log']:
            for p_class, mpmfs in {'total': all_mpmfs, ' metabolic': metab_mpmfs,
                                   ' transport': tx_mpmfs, ' processes ': pm_mpmfs}.items():
                if len(mpmfs) > 0:
                    x, y = (mpmfs[:, 0], mpmfs[:, 1]) if scale == 'lin' else self.get_log10_xy(mpmfs)
                    r_value, p_value = scipy.stats.pearsonr(x, y)
                    print(f'{p_class:16s}: r\N{SUPERSCRIPT TWO} = {r_value ** 2:.4f}, p = {p_value:.2e} '
                          f'({len(x):4d} proteins {scale} scale)')

    def save_to_escher(self, results_table, base_fname):
        """Export optimization results for use in Escher maps.

        Predicted reaction fluxes, protein concentrations or metabolite concentrations
        can be exported. For each condition in the table a separate export file is generated.
        Values with absolute value (<1e-8) are not exported.

        :param results_table: table with optimization results
        :type results_table: pandas.DataFrame or pandas.Series
        :param str base_fname: base name used to construct filenames
        """
        # in case a single condidion is selected, df could be a pandas Series
        if isinstance(results_table, pd.Series):
            results_table = pd.DataFrame(results_table)
        assert isinstance(results_table, pd.DataFrame)

        data_sets = {'rid': 'reaction data', 'gene': 'gene data', 'sid': 'metabolite data'}
        data_set = data_sets.get(results_table.index.name)

        count = 0
        if data_set:
            for condition in set(results_table.columns).intersection(self.results.keys()):
                values_dict = results_table[condition][abs(results_table[condition]) > 1e-8].to_dict()
                with open(f'{base_fname}_{condition}_{re.sub(" ", "_", data_set)}.json', 'w') as file:
                    json.dump(values_dict, file)
                    count += 1
            print(f'{count} file(s) exported for "Load {data_set}" into Escher maps')

        else:
            print(f'{results_table.index.name} not supported, use dataframe indexed with any of {data_sets.keys()}')

    # PLOT SUPPORT
    @ staticmethod
    def get_log10_xy(xy, cutoff=1e-6):
        """Return vectors log10(x) and log10(y) of 2D numpy array with two non-negative values per row

        Drop rows where x or y are below the cutoff (e.g. 1e-6)

        :param xy: 2D array with two float values per row (e.g. predicted vs. experimental values)
        :type: numpy.ndarray 2D: with two non-negative values per row
        :param float cutoff: min values of both x and y to return log10 values (default: 1e-6)
        :return: two vectors with log10(x) and log10(y) for rows that were not dropped
        :rtype: two numpy.ndarray with float values
        """
        log10_x = []
        log10_y = []
        for x, y in xy:
            if x > cutoff and y > cutoff:
                log10_x.append(np.log10(x))
                log10_y.append(np.log10(y))
        return np.array(log10_x), np.array(log10_y)

    def plot_grs(self, exp_grs, gr_max=None, highlight=None, plot_fname=None):
        """Plot predicted vs. measured growth rates.

        Using Matplotlib, plot correlation of predicted vs. measured growth rates.

        :param exp_grs: (media) conditions with measured growth rates (h-1)
        :type exp_grs: dict (key: condition/str, val: growth rate)
        :param float gr_max: max growth rate on axis (default: None)
        :param str highlight: condition to be highlighted (default: None)
        :param str plot_fname: filename to export plot (.pdf) (default: None)
        """
        marker2 = mpl.markers.MarkerStyle('o', fillstyle='full')

        sigma = self.optim.avg_enz_saturation
        conds = list(set(self.results.keys()).intersection(set(exp_grs.keys())))
        cond_exp_grs = [exp_grs[cond] for cond in conds]
        cond_pred_grs = [self.results[cond].objective_value for cond in conds]
        gr_max = gr_max if gr_max else max(max(cond_pred_grs), max(cond_exp_grs)) * 1.15

        fig, axs = plt.subplots(1, 1, figsize=(4.0, 4.0 * .618), squeeze=False)
        ax = axs[0, 0]

        ax.scatter(cond_exp_grs, cond_pred_grs, marker=marker2)
        if highlight and (highlight in conds):
            ax.scatter(exp_grs[highlight], self.results[highlight].objective_value)

        slope, intercept, rvalue, pvalue, stderr = scipy.stats.linregress(cond_exp_grs, cond_pred_grs)
        ax.plot((0.0, gr_max), (intercept, slope * gr_max + intercept), 'r:', lw=1)

        stats = r'$R^2$' + f'={rvalue ** 2:.4f}, p={pvalue:.2e}'
        ax.text(0.5, 0.1, stats, transform=ax.transAxes, va='top', ha='center')
        ax.text(0.01, 0.99, self.optim.model_name, transform=ax.transAxes, va='top')
        ax.text(0.01, 0.91, f'(saturation={sigma * 100:.1f}%)', transform=ax.transAxes, va='top')

        ax.set(xlim=(0.0, gr_max), ylim=(0.0, gr_max))
        ax.plot((0.0, 1.0), (0.0, 1.0), 'k--', lw=0.5)
        ax.set_xlabel(r'experimental growth rate ($h^{-1}$)')
        ax.set_ylabel(r'predicted growth rate ($h^{-1}$)')

        if plot_fname:
            fig.savefig(plot_fname)
        plt.show()

    def plot_proteins(self, condition, lin_max=None, plot_fname=None):
        """Plot predicted vs. measured protein levels for given condition.

        :param str condition: (media) condition in optimization results
        :param lin_max: max value in linear scale plot (default: None)
        :param str plot_fname: filename to export plot (.pdf) (default: None)
        """
        if hasattr(self, 'df_mpmf') is False:
            print('no proteomics data supplied for comparisson')
            return

        marker2 = mpl.markers.MarkerStyle('o', fillstyle='full')
        log_delta = np.log10(5.0)
        sigma = self.optim.avg_enz_saturation

        metab_gene2mpmfs = self.get_rtype_condition_mpmf('metabolic', condition)
        tx_gene2mpmfs = self.get_rtype_condition_mpmf('transport', condition)
        pm_gene2mpmfs = self.get_rtype_condition_mpmf('process', condition)
        all_gene2mpmfs = metab_gene2mpmfs | tx_gene2mpmfs | pm_gene2mpmfs

        metab_mpmfs = np.array(list(metab_gene2mpmfs.values()))
        tx_mpmfs = np.array(list(tx_gene2mpmfs.values()))
        pm_mpmfs = np.array(list(pm_gene2mpmfs.values()))
        all_mpmfs = np.array(list(all_gene2mpmfs.values()))

        fig, axs = plt.subplots(1, 2, figsize=(8.0, 4.0 * .618), squeeze=False)
        for gcol in [0, 1]:
            ax = axs[0, gcol]
            if gcol == 0:  # lin scale
                max_lin_val = 0.0
                for label, mpmfs in {'metabolic': metab_mpmfs, 'transport': tx_mpmfs, 'process': pm_mpmfs}.items():
                    if len(mpmfs) > 0:
                        ax.scatter(mpmfs[:, 0], mpmfs[:, 1], marker=marker2, label=label)
                        max_lin_val = max(max_lin_val, np.max(mpmfs))
                if lin_max is None:
                    lin_max = max_lin_val + 10.0
                xy_range = (0.0, lin_max)
                ax.set_xlabel(r'experimental mpmf (mg/gP)')
                ax.set_ylabel(r'predicted mpmf (mg/gP)')
                ax.legend(loc='lower right')
                ax.text(0.1, 0.99, self.optim.model_name, transform=ax.transAxes, va='top')
                ax.text(0.1, 0.91, f'(saturation={sigma * 100:.1f}%)', transform=ax.transAxes, va='top')
                ax.text(0.99, 0.5, f'[... {max_lin_val:.1f}]', transform=ax.transAxes, va='top', ha='right')

            else:  # log10 scale
                xy_range = (-7.0, 3.0)
                max_log_val = -10.0
                min_log_val = 10.0
                for label, mpmfs in {'metabolic': metab_mpmfs, 'transport': tx_mpmfs, 'process': pm_mpmfs}.items():
                    if len(mpmfs) > 0:
                        log_x, log_y = self.get_log10_xy(mpmfs)
                        ax.scatter(log_x, log_y, marker=marker2, label=label)
                        max_log_val = max(max_log_val, np.max(log_x), np.max(log_y))
                        min_log_val = min(min_log_val, np.min(log_x), np.min(log_y))

                log_x, log_y = self.get_log10_xy(all_mpmfs)
                slope, intercept, rvalue, pvalue, stderr = scipy.stats.linregress(log_x, log_y)
                ax.plot((xy_range[0], xy_range[1]),
                        (slope * xy_range[0] + intercept, slope * xy_range[1] + intercept), 'r:', lw=1)

                ax.plot((xy_range[0] + log_delta, xy_range[1]), (xy_range[0], xy_range[1] - log_delta), 'k:', lw=0.5)
                ax.plot((xy_range[0], xy_range[1] - log_delta), (xy_range[0] + log_delta, xy_range[1]), 'k:', lw=0.5)
                ax.text(0.99, 0.1, f'[{min_log_val:.1f} ... {max_log_val:.1f}]', transform=ax.transAxes,
                        va='top', ha='right')
                ax.set_xlabel(r'experimental mpmf log10')
                ax.set_ylabel(r'predicted mpmf log10')
                ax.legend(loc='upper left')

            ax.set(xlim=xy_range, ylim=xy_range)
            ax.plot(xy_range, xy_range, 'k--', lw=0.5)

        if plot_fname:
            fig.savefig(plot_fname)
        plt.show()



"""Implementation of EcmResults class.

Support functions for CobraPy and Gurobi optimization of EC (enzyme constraint) models
that have been created using the f2xba modelling package.

Support of GECKO, ccFBA, MOMENTmr and MOMENT optimization, as well
support for thermodynamic enhance models (TGECKO, TccFBA, TMOMENTmr, TMOMENT)

Peter Schubert, HHU Duesseldorf, CCB, April 2024
"""

import re
import pandas as pd
from collections import defaultdict

import f2xba.prefixes as pf
from .results import Results


class EcmResults(Results):
    """Optimization results processing for enzyme constraint models generated by f2xba.

    Processing and presentation of optimization results for a single or several
    (media) conditions.
    """


    def __init__(self, optim, results, df_mpmf=None):
        """Instantiate the EcmResults instance.

        :param optim: EcmOptimization instance
        :type optim: :class:`EcmOptimization`
        :param results: conditions with optimization results
        :type results: dict (key: str, val: :class:`Solution`)
        :param df_mpmf: measured protein mass fractions for conditions (default: None)
        :type df_mpmf: pandas.DataFrame
        """
        super().__init__(optim, results, df_mpmf)

        # Names of protein concentration variables are based on UniProt ids
        # These variables have an attached gene-product-reaction association with a gene product
        # The gene product mapps to the corresponding gene label
        self.uid2gene = {}
        for rid, row in self.optim.m_dict['reactions'].iterrows():
            if re.match(pf.V_PC_, rid) and type(row['fbcGeneProdAssoc']) is str:
                uid = re.sub(f'^{pf.V_PC_}', '', rid)
                gpid = re.sub('^assoc=', '', row['fbcGeneProdAssoc'])
                if gpid in self.optim.m_dict['fbcGeneProducts'].index:
                    gp_data = self.optim.m_dict['fbcGeneProducts'].loc[gpid]
                    self.uid2gene[uid] = [gp_data['label'], gp_data.get('name')]

    def get_predicted_protein_data(self, solution):
        """Extract protein mass wrt to total protein mass from single solution

        mpmf: 1000.0 * protein mass fraction i.e. (mg_protein/g_total_protein)

        Protein concentration variables are in mg/gDW
        total_protein is a constraint imposed by an upper bound
        Note: total cellular protein mass is part of cell_dry_weight, e.g. 57%
                  (other major parts would be RNA, lipids, DNA)
              total modeled protein mass is part of total cellular protein mass, e.g. 52.5%

        :param solution: optimization solution
        :type solution: :class:`Solution`
        :return: table with predicted protein concentrations
        :rtype: pandas.DataFrame
        """
        prot_data = []
        for rid, mg_active_per_gdw in solution.fluxes.items():
            if re.match(pf.V_PC_, rid) and rid != pf.V_PC_total:
                uid = re.sub(f'^{pf.V_PC_}', '', rid)
                label, gene_name = self.uid2gene[uid] if uid in self.uid2gene else [None, None]
                mpmf = mg_active_per_gdw / self.optim.protein_per_gdw
                prot_data.append([label, uid, gene_name, mpmf])
        cols = ['gene', 'uniprot', 'gene_name', 'mg_per_gP']
        df_prot_data = pd.DataFrame(prot_data, columns=cols).set_index('gene')
        return df_prot_data

    def collect_protein_results(self, units='mg_per_gP'):
        """Extract predicted protein concentrations from optimization results.

        Information columns are incorporated to provide context and facilitate data filtration.

        :param str units: concentration units: 'mg_per_gP' (default: 'mg_per_gP')
        :return: table with predicted proteins concentrations
        :rtype: pandas.DataFrame
        """
        supported_units = {'mg_per_gP'}
        if units not in supported_units:
            print(f'{units} not support, select any of {supported_units}')
            return pd.DataFrame()
        else:
            return self._collect_protein_results(units)

    def get_fluxes(self, solution):
        """Collect metabolic isoreaction fluxes for a single optimization solution.

        Excluding variables starting with 'V_', and exclude ARM reaction fluxes.

        :param solution: optimization solution
        :type solution: :class:`Solution`
        :return: reaction fluxes with additional reaction data
        :rtype: pandas.DataFrame
        """
        fluxes = {}
        for rid, mmol_per_gdwh in solution.fluxes.items():
            if re.match(pf.V_, rid) is None and re.search('_arm', rid) is None:
                rdata = self.optim.rdata[rid]
                fluxes[rid] = [rdata['reaction_str'], rdata['net_rid'], rdata['gpr'],
                               mmol_per_gdwh, abs(mmol_per_gdwh)]
        cols = ['reaction_str', 'net_rid', 'gpr', 'mmol_per_gDWh', 'abs mmol_per_gDWh']
        df_fluxes = pd.DataFrame(fluxes.values(), index=list(fluxes), columns=cols)
        df_fluxes.index.name = 'reaction'
        return df_fluxes

    def get_net_fluxes(self, solution):
        """net fluxes of a solution"""
        net_fluxes = defaultdict(float)
        for rid, val in solution.fluxes.items():
            if re.match(pf.V_, rid) is None and re.search('_arm', rid) is None:
                fwd_rid = re.sub('_REV$', '', rid)
                net_rid = re.sub(r'_iso\d*', '', fwd_rid)
                if re.search('_REV', rid):
                    net_fluxes[net_rid] -= val
                else:
                    net_fluxes[net_rid] += val

        # add reaction string and gene product relations
        net_flux_data = {}
        for rid, mmol_per_gdwh in net_fluxes.items():
            rdata = self.optim.net_rdata.get(rid)
            if rdata:
                net_flux_data[rid] = [rdata['reaction_str'], rdata['gpr'], mmol_per_gdwh, abs(mmol_per_gdwh)]
            else:
                net_flux_data[rid] = [None, None, mmol_per_gdwh, abs(mmol_per_gdwh)]

        cols = ['reaction_str', 'gpr', 'mmol_per_gDWh', 'abs mmol_per_gDWh']
        df_net_fluxes = pd.DataFrame(net_flux_data.values(), index=list(net_flux_data), columns=cols)
        df_net_fluxes.index.name = 'rid'
        return df_net_fluxes



"""Implementation of SbmlParameter class.

based on XBAparameter from xbanalysis package

Peter Schubert, HHU Duesseldorf, February 2023
"""

from .sbml_sbase import SbmlSBase


class SbmlParameter(SbmlSBase):

    def __init__(self, s_parameter):
        """Instantiate SbmlParameter instance with data from s_parameter

        s_parameter is pandas Series with Series.name attribute set parameter id and
        several mandatory and optional attributes based on SBML specifications.

        - mandatory attributes:
            - Series.name: str - parameter id
            - 'value': float
            - 'constant': bool
            - 'units': str - unit definition id

        - optional attributes:
            - 'name': str - handled in parent class
            - 'sboterm': str - handled in parent class
            - 'metaid': str - handled in parent class
            - 'miriamAnnotation': str - handled in parent class
            - 'notes': str - handled in parent class

        :param s_parameter: parameter data from SBML import
        :type s_parameter: pandas.Series
        """
        super().__init__(s_parameter)
        self.value = s_parameter['value']
        self.constant = s_parameter.get('constant', True)
        self.units = s_parameter.get('units', 'dimensionless')
        self.reuse = True

    def to_dict(self):
        data = super().to_dict()
        data['value'] = self.value
        data['constant'] = self.constant
        data['units'] = self.units
        return data

    def modify_attribute(self, attribute, value):
        """Modify attribute value.

        :param str attribute: attribute name
        :param value: value to be configured
        :type value: str or bool
        """
        if hasattr(self, attribute):
            setattr(self, attribute, value)
        else:
            print(f'unknown parameter attribute {attribute}')



"""Implementation of SbmlReaction class.

based on XbaReaction from xbanalysis package

Peter Schubert, HHU Duesseldorf, February 2023
"""

import re
import numpy as np

import sbmlxdf
from .sbml_sbase import SbmlSBase


class SbmlReaction(SbmlSBase):

    def __init__(self, s_reaction, species_dict):
        """Instantiate SbmlReaction instance with data from s_reaction

        s_reaction is pandas Series with Series.name attribute set reaction id and
        several mandatory and optional attributes based on SBML specifications.

        - mandatory attributes:
            - Series.name: str - reaction id
            - 'reversible': bool
            - 'reactants': str 'species=<sid>, stoic=float; ...', or dict {<sid>: float, ...}
            - 'products': str, or dict
            - 'fbcLowerFluxBound': str - parameter id
            - 'fbcUpperFluxBound': str - parameter id

        - optional attributes:
            - 'name': str - handled in parent class
            - 'sboterm': str - handled in parent class
            - 'metaid': str - handled in parent class
            - 'miriamAnnotation': str - handled in parent class
            - 'notes': str - handled in parent class
            - 'fbcGeneProdAssoc': str
            - 'kind': str

        A reaction compartment is set based on compartments of reaction reactants/products.
        In case of transporters, the compartment ids are concatenated using '-'

        :param s_reaction: reaction data from SBML import
        :type s_reaction: pandas Series
        :param species_dict: Species configured in the model to extract compartment info
        :type species_dict: dict (key: sid, val: Class SBMLspecies)
        """
        super().__init__(s_reaction)

        self.reversible = s_reaction['reversible']
        self.reactants = s_reaction['reactants']
        self.products = s_reaction['products']
        self.sref2id = s_reaction.get('sref2id', {})

        self.fbc_lower_bound = s_reaction['fbcLowerFluxBound']
        self.fbc_upper_bound = s_reaction['fbcUpperFluxBound']
        self.gene_product_assoc = s_reaction.get('fbcGeneProdAssoc')
        self.scale = s_reaction.get('scale', 1.0)

        self.rp_counts = [len(self.reactants), len(self.products)]
        self.rp_compartments = [self.get_compartments(self.reactants, species_dict),
                                self.get_compartments(self.products, species_dict)]
        self.compartment = '-'.join(sorted(self.rp_compartments[0].union(self.rp_compartments[1])))
        self.kind = s_reaction.get('kind', self.get_reaction_kind())
        self.enzymes = []
        self.kcatsf = None
        self.kcatsr = None
        self.orig_rid = self.id

    @property
    def ec_codes(self):
        return self.miriam_annotation.get_qualified_refs('bqbiol:is', 'ec-code')

    @property
    def reactants(self):
        return self._reactants

    @reactants.setter
    def reactants(self, value):
        if type(value) is dict:
            self._reactants = value.copy()
        else:
            self._reactants = self.get_srefs(value)

    @property
    def products(self):
        return self._products

    @products.setter
    def products(self, value):
        if type(value) is dict:
            self._products = value.copy()
        else:
            self._products = self.get_srefs(value)

    @property
    def reaction_string(self):
        return self.get_reaction_string()

    @property
    def gene_product_assoc(self):
        return self._gene_product_assoc

    @gene_product_assoc.setter
    def gene_product_assoc(self, value):
        if type(value) is str and len(value) > 1:
            self._gene_product_assoc = re.sub(r'^assoc=', '', value)
        else:
            self._gene_product_assoc = None

    def modify_attribute(self, attribute, value):
        """modify attribute value.

        Special case for reactants and products, where individual
        species reference can be deleted or updated individually using
        'reactant', 'product' attributes.

        :param attribute: attribute name
        :type attribute: str
        :param value: value to be configured
        :type value: str
        """
        if attribute in {'reactant', 'product'}:
            sid, _val = value.split('=')
            val = float(_val)
            if attribute == 'reactant':
                if val == 0.0:
                    del self.reactants[sid]
                else:
                    self.reactants[sid] = val
            else:
                if val == 0.0:
                    del self.products[sid]
                else:
                    self.products[sid] = val
        else:
            setattr(self, attribute, value)

    @staticmethod
    def get_srefs(srefs_str):
        """Extract composition from srefs string (component and stoichiometry).

        Species references string contains ';' separated records of composition.
        Each record contains ',' separated key=value pairs. Required keys are
        'species' and 'stoic'.

        :param srefs_str: species references string with attibutes 'species' and 'stoic'
        :type srefs_str: str
        :return: composition (components with stoichiometry
        :rtype: dict (key: species id, value: stoichiometry (float)
        """
        srefs = {}
        if type(srefs_str) == str:
            for sref_str in sbmlxdf.record_generator(srefs_str):
                params = sbmlxdf.extract_params(sref_str)
                srefs[params['species']] = float(params['stoic'])
        return srefs

    @staticmethod
    def get_compartments(sids, species_dict):
        """Retrieve compartments of species

        :param sids: species
        :type sids: list of str
        :param species_dict: Species configured in the model
        :type species_dict: dict (key: sid, val: Class SBMLspecies)
        """
        compartments = set()
        for sid in sids:
            compartments.add(species_dict[sid].compartment)
        return compartments

    def get_reaction_kind(self):
        """determine the kind/type of reaction.

        Notes:
            - sink reactions get assigned 'exchange' kind
            - biomass reactions usually get assigned 'transporter'
            - sink and biomass reactions are not catalyzed by enzymes

        :return: kind of the reaction
        :rtype: str
        """
        kind = 'metabolic'

        if (self.rp_counts[0] == 0) or (self.rp_counts[1] == 0):
            kind = 'exchange'
        elif len(self.rp_compartments[0].union(self.rp_compartments[1])) > 1:
            kind = 'transporter'
        return kind

    def to_dict(self):
        data = super().to_dict()
        data['reversible'] = self.reversible
        data['fbcLowerFluxBound'] = self.fbc_lower_bound
        data['fbcUpperFluxBound'] = self.fbc_upper_bound

        srefs = []
        for sid, stoic in self.reactants.items():
            sref_id_str = f'id={self.sref2id[sid]}, ' if sid in self.sref2id else ''
            srefs.append(f'{sref_id_str}species={sid}, stoic={stoic}, const=True')
        data['reactants'] = '; '.join(srefs)
        srefs = []
        for sid, stoic in self.products.items():
            sref_id_str = f'id={self.sref2id[sid]}, ' if sid in self.sref2id else ''
            srefs.append(f'{sref_id_str}species={sid}, stoic={stoic}, const=True')
        data['products'] = '; '.join(srefs)

        if self.gene_product_assoc:
            data['fbcGeneProdAssoc'] = f'assoc={self.gene_product_assoc}'
        return data

    def gpa_remove_gps(self, del_gps):
        """Remove gene products from Gene Product Rules.

        Used to remove dummy protein gene product and
        gene products related to coezymes that already
        appear in reaction reactants/products

        :param del_gps: coenzyme gene products
        :type del_gps: set or list of str
        """
        if self.gene_product_assoc:
            gpa = self.gene_product_assoc
            for gp in del_gps:
                if gp in self.gene_product_assoc:
                    gpa = re.sub(gp + ' and ', '', gpa)
                    gpa = re.sub(' and ' + gp, '', gpa)
                    gpa = re.sub(gp + ' or ', '', gpa)
                    gpa = re.sub(' or ' + gp, '', gpa)
                    gpa = re.sub(gp, '', gpa)
            # if all gene_products have been removed, set gpa to None
            if len(set(re.findall(r"\w+", gpa)).difference({'or', 'and'})) > 0:
                self.gene_product_assoc = gpa
            else:
                self.gene_product_assoc = None

    def set_enzymes(self, eids):
        """Add enzymes to the reaction, also kcat arrays.

        :param eids: enzymes ids to set
        :type eids: list of str
        """
        self.enzymes = sorted(eids)
        if len(self.enzymes) > 0:
            self.kcatsf = np.full(len(self.enzymes), np.nan)
            if self.reversible is True:
                self.kcatsr = np.full(len(self.enzymes), np.nan)

    def set_kcat(self, eid, dirxn, kcat):
        """Set kcat value for specified enzyme and reaction direction.

        We only add kcat values for enzymes and direction that are valid.

        in case enzyme id is set to 'unspec', configure kcats value for given
         direction on all isoenzymes

        :param eid: enzyme id, e.g. 'enz_b0007', used 'unspec' to set kcat for all
        :type eid: str
        :param dirxn: reaction direction (1: forward, -1: reverse)
        :type dirxn: int or float {-1, 1}
        :param kcat: catalytic rate for enzyme catalyzed reaction in per second
        :type kcat: float (non-negative values only)
        :return: number of successful updates
        :rtype: int
        """
        assert(abs(dirxn) == 1)
        assert(kcat >= 0.0)

        n_updates = 0
        kcats_dir = 'kcatsf' if dirxn == 1 else 'kcatsr'
        if getattr(self, kcats_dir) is not None:
            if eid in self.enzymes:
                idx = self.enzymes.index(eid)
                getattr(self, kcats_dir)[idx] = kcat
                n_updates = 1
            else:
                if eid == 'unspec':
                    for idx in range(len(self.enzymes)):
                        getattr(self, kcats_dir)[idx] = kcat
                        n_updates += 1
        return n_updates

    def scale_kcat(self, eid, dirxn, scale):
        """Scale kcat value for specified enzyme and reaction direction.

        We only scale kcat values for enzymes and direaction that are valid.

        in case enzyme id is set to 'unspec', all kcats for all isoenzymes are scaled

        :param eid: enzyme id, e.g. 'enz_b0007', used 'unspec' to set kcat for all
        :type eid: str
        :param dirxn: reaction direction (1: forward, -1: reverse)
        :type dirxn: int or float {-1, 1}
        :param scale: scale factor by which to divide existing kcat value
        :type scale: float (positive values only)
        :return: number of successful updates
        :rtype: int
        """
        assert(abs(dirxn) == 1)
        assert(scale > 0.0)

        n_updates = 0
        kcats_dir = 'kcatsf' if dirxn == 1 else 'kcatsr'
        if getattr(self, kcats_dir) is not None:
            if eid in self.enzymes:
                idx = self.enzymes.index(eid)
                ref_kcat = getattr(self, kcats_dir)[idx]
                getattr(self, kcats_dir)[idx] = ref_kcat / scale
                n_updates = 1
            else:
                if eid == 'unspec':
                    for idx in range(len(self.enzymes)):
                        ref_kcat = getattr(self, kcats_dir)[idx]
                        getattr(self, kcats_dir)[idx] = ref_kcat / scale
                        n_updates += 1
        return n_updates

    def modify_stoic(self, srefs):
        """Update reactants/products stoichiometry.

        stoic < 0: reactant
        stoic > 0: product
        stoic == None: remove species

        :param srefs: species with stoichiometry
        :type srefs: dict (key: sid, val: stoic, float or None)
        """
        for sid, stoic in srefs.items():
            if stoic <= 0.0:
                if sid in self.products:
                    del self.products[sid]
                if stoic < 0.0:
                    self.reactants[sid] = -stoic
            if stoic >= 0.0:
                if sid in self.reactants:
                    del self.reactants[sid]
                if stoic > 0.0:
                    self.products[sid] = stoic

    def modify_bounds(self, bound_pids):
        """Update flux bounds with new parameter id.

        :param bound_pids: bounds to be updated 'lb' or 'ub' with parameter ids
        :type bound_pids: dict (key: 'ub' and/or 'lb, val: parameter id
        """
        if 'lb' in bound_pids:
            self.fbc_lower_bound = bound_pids['lb']
        if 'ub' in bound_pids:
            self.fbc_upper_bound = bound_pids['ub']

    def reduce_reactants(self):
        """Reduce reactant/products appearing on both side of the reaction string
        """
        reduce_sids = set(self.reactants).intersection(self.products)
        for sid in reduce_sids:
            if self.reactants[sid] == self.products[sid]:
                del self.reactants[sid]
                del self.products[sid]
            elif self.reactants[sid] > self.products[sid]:
                self.reactants[sid] -= self.products[sid]
                del self.products[sid]
            else:
                self.products[sid] -= self.reactants[sid]
                del self.reactants[sid]

    def add_delta_product(self, sid, delta_stoic):
        """Modify reaction stoichiometry by adding metabolites to the product side.

        Alternatively reducing metabolites from the reactant side.
        delta_stoic can be positive or negative.

        :param sid: species id
        :type sid: str
        :param delta_stoic: stoichiomeric amount to increase product
        :type delta_stoic: float
        """
        if delta_stoic > 0.0:
            if sid in self.products:
                self.products[sid] += delta_stoic
            else:
                self.products[sid] = delta_stoic
                self.reduce_reactants()
        else:
            if sid in self.reactants:
                self.reactants[sid] += (-delta_stoic)
            else:
                self.reactants[sid] = (-delta_stoic)
                self.reduce_reactants()

    def set_gpa(self, gpa):
        """Change the gene product association.

        gpa with value '' will remove the gpa attribute

        :param gpa: gene product association
        :type gpa: str
        """
        self.gene_product_assoc = gpa
        if gpa == '':
            self.gene_product_assoc = None

    def correct_reversibility_old(self, parameters, exclude_ex_reactions=False):
        """Correct reversibility based on flux bounds.

        Note issues with reactions that have lb and ub both being negative (E.g. ATPase in MMSYN)

        reversible is set to False if flux range is either positive or negative.
        exchange reactions can be exluded

        :param parameters: flux bound parameters
        :type parameters: SbmlParameters
        :param exclude_ex_reactions: flag if exchange reactions should be excluded
        :type exclude_ex_reactions: bool, (default: False)
        """
        lb_value = parameters[self.fbc_lower_bound].value
        ub_value = parameters[self.fbc_upper_bound].value
        if self.reversible is True and (lb_value >= 0.0 or ub_value <= 0.0):
            if exclude_ex_reactions is False:
                self.reversible = False
            else:
                if len(self.products) > 0:
                    self.reversible = False

    def replace_sids(self, old2newsid):
        """replace species ids in reactants/products based on supplied mapping

        :params old2newsid: mapping from old to new species ids
        :type old2newsid: dict (key: old sid, val: new sid)
        """
        for oldsid in list(self.reactants):
            if oldsid in old2newsid:
                newsid = old2newsid[oldsid]
                self.reactants[newsid] = self.reactants[oldsid]
                del self.reactants[oldsid]

        for oldsid in list(self.products):
            if oldsid in old2newsid:
                newsid = old2newsid[oldsid]
                self.products[newsid] = self.products[oldsid]
                del self.products[oldsid]

    def reverse_substrates(self):
        reactants = self.products
        self.products = self.reactants
        self.reactants = reactants

    def set_id(self, rid):
        self.id = rid
        if hasattr(self, 'metaid'):
            self.metaid = f'meta_{rid}'

    def get_reaction_string(self):
        """Generate a reaction string from reactants/products/reversible.

        :return: reaction string
        :rtype: str
        """
        lparts = []
        for sid, stoic in self.reactants.items():
            stoic_str = str(stoic) + ' ' if stoic != 1.0 else ''
            lparts.append(f'{stoic_str}{sid}')
        left_side = ' + '.join(lparts)
        rparts = []
        for sid, stoic in self.products.items():
            stoic_str = str(stoic) + ' ' if stoic != 1.0 else ''
            rparts.append(f'{stoic_str}{sid}')
        right_side = ' + '.join(rparts)

        direction = ' -> ' if self.reversible is True else ' => '

        return left_side + direction + right_side



"""Implementation of RbaMetabolism, RbaComponent, RbaSpecies and RbaReaction classes.

Peter Schubert, CCB, HHU Duesseldorf, December 2022
"""

import os
import pandas as pd
from xml.etree.ElementTree import parse, ElementTree, Element, SubElement, indent

from ..utils.rba_utils import get_species_refs_from_xml, get_species_refs_from_str


class RbaMetabolism:

    def __init__(self):
        self.compartments = {}
        self.species = {}
        self.reactions = {}

    def import_xml(self, model_dir):

        file_name = os.path.join(model_dir, 'metabolism.xml')
        if os.path.exists(file_name) is True:
            root = parse(file_name).getroot()
            assert root.tag == 'RBAMetabolism'
            self.compartments = RbaCompartment.import_xml(root.find('listOfCompartments'))
            self.species = RbaSpecies.import_xml(root.find('listOfSpecies'))
            self.reactions = RbaReaction.import_xml(root.find('listOfReactions'))
        else:
            print(f'{file_name} not found!')

    def from_df(self, m_dict):
        if 'compartments' in m_dict:
            self.compartments = RbaCompartment.from_df(m_dict['compartments'])
        else:
            print(f'compartments not imported!')
        if 'species' in m_dict:
            self.species = RbaSpecies.from_df(m_dict['species'])
        else:
            print(f'species not imported!')
        if 'reactions' in m_dict:
            self.reactions = RbaReaction.from_df(m_dict['reactions'])
        else:
            print(f'reactions not imported!')

        if 'compartments' in m_dict:
            self.compartments = RbaCompartment.from_df(m_dict['compartments'])
        if 'species' in m_dict:
            self.species = RbaSpecies.from_df(m_dict['species'])
        if 'reactions' in m_dict:
            self.reactions = RbaReaction.from_df(m_dict['reactions'])

    def from_xba(self, rba_params, xba_model):
        """Add compartments, species and reactions from XbaModel and RBA specific parameters.

        Compartment data is retrieved from rba_params['compartments']
        Species and Reactions data is retrieved from xba_model

        :param rba_params: RBA model specific parametrization
        :type rba_params: dict of pandas DataFrames
        :param xba_model: xba model based on genome scale metabolic model
        :type xba_model: Class XbaModel
        """
        for cid, row in rba_params['compartments'].iterrows():
            self.compartments[cid] = RbaCompartment(cid, row['name'])
        print(f'{len(self.compartments):4d} compartments')

        for sid, s in xba_model.species.items():
            boundary_condition = True if s.compartment == xba_model.external_compartment else False
            self.species[sid] = RbaSpecies(sid, boundary_condition)
        print(f'{len(self.species):4d} species')

        for rid, r in xba_model.reactions.items():
            if r.kind != 'exchange' and r.kind != 'biomass':
                self.reactions[rid] = RbaReaction(rid, r.reversible, r.reactants.copy(), r.products.copy())
        print(f'{len(self.reactions):4d} reactions')

    def export_xml(self, model_dir):
        file_name = os.path.join(model_dir, 'metabolism.xml')
        root = Element('RBAMetabolism')

        compartments = SubElement(root, 'listOfCompartments')
        for item in self.compartments.values():
            compartments.append(item.export_xml())

        species = SubElement(root, 'listOfSpecies')
        for item in self.species.values():
            species.append(item.export_xml())

        reactions = SubElement(root, 'listOfReactions')
        for item in self.reactions.values():
            reactions.append(item.export_xml())

        tree = ElementTree(root)
        indent(tree)
        tree.write(file_name)

    def to_df(self):
        df_c = pd.DataFrame([item.to_dict() for item in self.compartments.values()])
        df_c.index.name = 'index'
        df_s = pd.DataFrame([item.to_dict() for item in self.species.values()])
        df_s.set_index('species', inplace=True)
        df_r = pd.DataFrame([item.to_dict() for item in self.reactions.values()])
        df_r.set_index('reaction', inplace=True)
        return {'compartments': df_c, 'species': df_s, 'reactions': df_r}

    def validate(self, component_ids):
        valid = True
        missing = self.ref_molecules().difference(component_ids['species'])
        if len(missing) > 0:
            print('species used in reactions not defined:', missing)
            valid = False
        return valid

    def ref_molecules(self):
        refs = set()
        for r in self.reactions.values():
            refs |= {sid for sid in r.reactants}
            refs |= {sid for sid in r.products}
        return refs


class RbaCompartment:

    def __init__(self, cid, name=None):
        self.id = cid
        self.name = name

    @staticmethod
    def import_xml(compartments):
        data = {}
        for compartment in compartments.findall('compartment'):
            cid = compartment.attrib['id']
            rba_compartment = RbaCompartment(cid)
            data[cid] = rba_compartment
        return data

    @staticmethod
    def from_df(df):
        data = {}
        for _, row in df.iterrows():
            cid = row['compartment']
            rba_compartment = RbaCompartment(cid)
            data[cid] = rba_compartment
        return data

    def export_xml(self):
        return Element('compartment', {'id': self.id})

    def to_dict(self):
        return {'compartment': self.id}


class RbaSpecies:

    def __init__(self, sid, boundary_condition=False):
        self.id = sid
        self.boundary_condition = boundary_condition

    @staticmethod
    def import_xml(species):

        data = {}
        for sp in species.findall('species'):
            sid = sp.attrib['id']
            rba_species = RbaSpecies(sid)
            if sp.attrib['boundaryCondition'].lower() == 'true':
                rba_species.boundary_condition = True
            data[sid] = rba_species
        return data

    @staticmethod
    def from_df(df):
        data = {}
        for sid, row in df.iterrows():
            rba_species = RbaSpecies(sid)
            rba_species.boundary_condition = row['boundaryCondition']
            data[sid] = rba_species
        return data

    def export_xml(self):
        attribs = {'id': self.id, 'boundaryCondition':  str(self.boundary_condition).lower()}
        return Element('species', attribs)

    def to_dict(self):
        return {'species': self.id, 'boundaryCondition': self.boundary_condition}


class RbaReaction:

    def __init__(self, sid, reversible=True, reactants=None, products=None):
        self.id = sid
        self.reversible = reversible
        self.reactants = reactants if reactants is not None else None
        self.products = products if products is not None else None

    @staticmethod
    def import_xml(reactions):

        data = {}
        for reaction in reactions.findall('reaction'):
            rid = reaction.attrib['id']
            rba_reaction = RbaReaction(rid)
            if reaction.attrib['reversible'].lower() == 'false':
                rba_reaction.reversible = False
            rba_reaction.reactants = get_species_refs_from_xml(reaction.find('listOfReactants'))
            rba_reaction.products = get_species_refs_from_xml(reaction.find('listOfProducts'))
            data[rid] = rba_reaction
        return data

    @staticmethod
    def from_df(df):
        data = {}
        for rid, row in df.iterrows():
            rba_reaction = RbaReaction(rid)
            rba_reaction.reversible = row['reversible']
            rba_reaction.reactants = get_species_refs_from_str(row['reactants'])
            rba_reaction.products = get_species_refs_from_str(row['products'])
            data[rid] = rba_reaction
        return data

    def export_xml(self):
        attribs = {'id': self.id, 'reversible': str(self.reversible).lower()}
        reaction = Element('reaction', attribs)

        reactants = SubElement(reaction, 'listOfReactants')
        for sid, stoic in self.reactants.items():
            SubElement(reactants, 'speciesReference', {'species': sid, 'stoichiometry': str(stoic)})

        products = SubElement(reaction, 'listOfProducts')
        for sid, stoic in self.products.items():
            SubElement(products, 'speciesReference', {'species': sid, 'stoichiometry': str(stoic)})
        return reaction

    def to_dict(self):
        reactants = '; '.join([f'species={species}, stoic={stoic}'
                               for species, stoic in self.reactants.items()])
        products = '; '.join([f'species={species}, stoic={stoic}'
                              for species, stoic in self.products.items()])
        return {'reaction': self.id, 'reversible': self.reversible, 'reactants': reactants, 'products': products}



"""Implementation of EcmOptimization class.

Support functions for cobrapy and gurobipy optimization of EC (enzyme constraint) models
that have been created using the f2xba modelling package.

Support of GECKO, ccFBA, MOMENTmr and MOMENT optimization, as well
support for thermodynamic enhance models (TGECKO, TccFBA, TMOMENTmr, TMOMENT)

Peter Schubert, HHU Duesseldorf, CCB, April 2024
"""

import re
import numpy as np
import pandas as pd
from collections import defaultdict
# Gurobipy should not be a hard requirement, unless used in this context
try:
    import gurobipy as gp
except ImportError:
    gp = None
    pass

import f2xba.prefixes as pf
from .optimize import Optimize, status2text


class EcmOptimization(Optimize):
    """Optimization support for enzyme constraint models.

    Optimization support is provided via cobrapy and guropibpy interfaces for
    enzyme constraint models created by f2xba. When utilizing cobrapy, it is
    necessary to first load the model using SBML.
    """

    def __init__(self, fname, cobra_model=None):
        """Instantiate the EcmOptimization instance.

        :param str fname: filename of the SBML coded extended model
        :param cobra_model: reference to a cobrapy model (default: None)
        :type cobra_model: cobra.Model
        """
        super().__init__('ECM', fname, cobra_model)
        self.orig_coupling = {}

        self.ecm_type = self.m_dict['modelAttrs'].get('id', '_GECKO').rsplit('_', 1)[1]
        if self.ecm_type.endswith('MOMENT'):
            self.configure_moment_model_constraints()

    def configure_moment_model_constraints(self):
        """Configure constraints related to MOMENT model optimization.

        Must be invoked after loading a MOMENT type model (not MOMENTmr). MOMENT implements promiscuous
        enzymes, that can catalyze alternative reactions without additional costs. The most costly
        reaction flux needs to be supported and all alternative reaction fluxes come for free.
        """
        if self.is_gpm:
            for constr in self.gpm.getConstrs():
                if re.match(pf.C_prot_, constr.ConstrName) and re.match(pf.C_prot_pool, constr.ConstrName) is None:
                    constr.sense = '>'
        else:
            for constr in self.model.constraints:
                if re.match(pf.C_prot_, constr.name) and re.match(pf.C_prot_pool, constr.name) is None:
                    constr.ub = 1000.0
        print(f'MOMENT protein constraints configured ≥ 0')

    def scale_kcats(self, scale_kcats):
        """Scale turnover number values for model tuning.

        Use unscale_kcats() to return to original values, e.g.

        .. code-block:: python

            eo = EcmOptimization('iJO1366_GECKO.xml)
            scale_kcats= {'BPNT': 0.2, 'TMPK':0.2, 'TMPK_REV':0.2, 'CLPNS':0.25}
            eo.scale_kcats(scale_kcats)
            solution = eo.optimize()
            eo.unscale_kcats()

        :param scale_kcats: reaction identifiers (without prefix `R_`) and scaling factor
        :type scale_kcats: dict (key: reaction id/str, val: factor/float)
        """
        if self.is_gpm:
            self._gp_scale_kcats(scale_kcats)
        else:
            self._cp_scale_kcats(scale_kcats)

    def unscale_kcats(self):
        """Reset previously scaled turnover numbers.
        """
        if self.is_gpm:
            self._gp_unscale_kcats()
        else:
            self._cp_unscale_kcats()

    def _cp_scale_kcats(self, scale_kcats):
        """Scale kcat values for COBRApy interface, by updating coupling coefficients.

        :param scale_kcats: selected reaction ids (without 'R_' prefix) with kcat scaling factor
        :type scale_kcats: dict (key: reaction id without 'R_' prefix, val: float)
        """
        assert self.is_gpm is False, 'applicable to COBRApy interface only'
        orig_coupling = defaultdict(dict)
        for ridx, scale in scale_kcats.items():
            assert ridx in self.model.reactions
            r = self.model.reactions.get_by_id(ridx)
            for m, coeff in r.metabolites.items():
                if re.match(f'{pf.C_prot_}', m.id):
                    r.add_metabolites({m: coeff / scale}, combine=False)
                    orig_coupling[ridx][m.id] = coeff
        self.orig_coupling = dict(orig_coupling)

    def _cp_unscale_kcats(self):
        """Unscale kcat values for COBRApy inteface, by resetting original coupling coefficients
        """
        assert self.is_gpm is False, 'applicable to COBRApy interface only'
        for ridx, couplings in self.orig_coupling.items():
            r = self.model.reactions.get_by_id(ridx)
            for constr_id, coeff in couplings.items():
                r.add_metabolites({constr_id: coeff}, combine=False)
        self.orig_coupling = {}

    def _gp_scale_kcats(self, scale_kcats):
        """Scale kcat values for Gurobi interface, by updating coupling coefficients.

        :param scale_kcats: selected reaction ids (without 'R_' prefix) with kcat scaling factor
        :type scale_kcats: dict (key: reaction id without 'R_' prefix, val: float)
        """
        assert self.is_gpm is True, 'applicable to Gurobi interface only'
        orig_coupling = defaultdict(dict)
        self.gpm.update()
        for ridx, scale in scale_kcats.items():
            var = self.gpm.getVarByName(f'{pf.R_}{ridx}')
            col = self.gpm.getCol(var)
            for idx in range(col.size()):
                constr = col.getConstr(idx)
                if re.match(f'{pf.C_prot_}', constr.getAttr('ConstrName')):
                    coeff = col.getCoeff(idx)
                    self.gpm.chgCoeff(constr, var, coeff / scale)
                    orig_coupling[ridx][constr.getAttr('ConstrName')] = coeff
        self.gpm.update()
        self.orig_coupling = dict(orig_coupling)

    def _gp_unscale_kcats(self):
        """reset kcat values to original values.

        Used in manually tuning model kcats
        - call gp_tune_scale_kcats() prior to optmization
        - call gp_tune_reset_kcats() after optimizatin to reset old kcat values
        """
        assert self.is_gpm is True, 'applicable to Gurobi interface only'
        for ridx, couplings in self.orig_coupling.items():
            var = self.gpm.getVarByName(f'{pf.R_}{ridx}')
            for constr_id, coeff in couplings.items():
                constr = self.gpm.getConstrByName(constr_id)
                self.gpm.chgCoeff(constr, var, coeff)
        self.gpm.update()
        self.orig_coeffs = {}

    def gp_fit_kcats(self, proteins_mg_per_gdw, max_scale_factor=1000.0):

        # create mapping of gene to uniprot id based on V_PC_xxxx variables
        gene2uid = {}
        for rid, row in self.m_dict['reactions'].iterrows():
            if re.match(pf.V_PC_, rid) and type(row['fbcGeneProdAssoc']) is str:
                uid = re.sub(f'^{pf.V_PC_}', '', rid)
                gpid = re.sub('^assoc=', '', row['fbcGeneProdAssoc'])
                if gpid in self.m_dict['fbcGeneProducts'].index:
                    gene = self.m_dict['fbcGeneProducts'].at[gpid, 'label']
                    gene2uid[gene] = uid

        # create slack model from original model
        self.gpm.update()
        slack_gpm = self.gpm.copy()

        # add slack variables (positive and negative)
        slack_vars = []
        for gene, mg_per_gdw in proteins_mg_per_gdw.items():
            uid = gene2uid[gene]

            # fix protein concentrations to proteomics values (selected proteins only)
            var = slack_gpm.getVarByName(f'{pf.V_PC_}{uid}')
            var.lb = mg_per_gdw
            var.ub = mg_per_gdw

            # create positive and negative slack variables for protein concentrations
            pslack_var = slack_gpm.addVar(lb=0.0, ub=100, vtype='C', name=f'{pf.V_PSLACK_}{uid}')
            nslack_var = slack_gpm.addVar(lb=0.0, ub=100, vtype='C', name=f'{pf.V_NSLACK_}{uid}')
            slack_vars.append(pslack_var)
            slack_vars.append(nslack_var)

            # couple pos/neg slack variable with protein concentration constraint
            constr = slack_gpm.getConstrByName(f'{pf.C_prot_}{uid}')
            slack_gpm.chgCoeff(constr, pslack_var, 1.0)
            slack_gpm.chgCoeff(constr, nslack_var, -1.0)

        # 1. optimize for maximum growth rate under protein constraint conditions
        slack_gpm.optimize()
        status = status2text[slack_gpm.status]
        if status.lower() != 'optimal':
            print(f'slack model optimization (max growth rate) returned status: {status}')
            return {}

        max_gr = slack_gpm.objval
        print(f'successful slack optimization with growth rate of {max_gr:.3f} h-1')

        # 2. minimize sums of slack variables to avoid loops
        # fix maximum growth rate by adding a new constraint
        model_objective = slack_gpm.getObjective()
        slack_gpm.addLConstr(model_objective, '=', max_gr, 'Max growth rate objective')

        # New Objective: minimize sum of all non-negative reactions
        slack_gpm.update()

        slack_objective = gp.LinExpr(np.ones(len(slack_vars)), slack_vars)
        slack_gpm.setObjective(slack_objective, gp.GRB.MINIMIZE)

        # optimize and collect results (without error handling)
        slack_gpm.optimize()
        status = status2text[slack_gpm.status]
        if status.lower() != 'optimal':
            print(f'slack model optimization (minimize sum of slack) returned status: {status}')
            return {}

        # process slack model results to determine kcat scaling factors
        gene2scale = {}
        for gene, exp_mg_per_gdw in proteins_mg_per_gdw.items():
            uid = gene2uid[gene]
            pos_slack = slack_gpm.getVarByName(f'{pf.V_PSLACK_}{uid}').x
            neg_slack = slack_gpm.getVarByName(f'{pf.V_NSLACK_}{uid}').x
            pred_mg_per_gdw = exp_mg_per_gdw + pos_slack - neg_slack

            scale = pred_mg_per_gdw / exp_mg_per_gdw
            scale = min(scale, max_scale_factor)
            # only capture postive >0 values for scale, i.e. predicted concentration > 0
            if pred_mg_per_gdw > 1e-8:
                gene2scale[gene] = scale

        min_scale = min(gene2scale.values())
        max_scale = max(gene2scale.values())
        print(f'{len(gene2scale)} genes to be scaled by factors of {min_scale:.4f} ... {max_scale:.3f}')
        return gene2scale

    @staticmethod
    def create_fitted_kcats_file(gene2scale, orig_kcats_fname, fitted_kcats_fname):
        """Scale original kcats for selected genes and create fitted kcats Excel document

        - genes products (proteins) can be part of one or several enzymes
        - enzyme may catalyze several reactions, including forward/reverse split isoreactions
        - for each selected gene with a kcat scaling factor, retrieve respective kcat records from original
          kcats Excel Document (one record per isoenzyme)
        - calculated fitted kcats for each kcat record where gene product is used.
        - in case of an enzyme complex consisting of several gene products used in an isoreaction, calculate a single
          kcat value, here we used geometric mean across all individual kcats for the kcat record
        - we also limit kcats to be between 1000 and 0.0001 s-1
        - configure scaled kcat records and store resulting fitted kcat Excel document

        :param gene2scale: gene labels (proteins) with kcat scaling factors
        :type gene2scale: dict (key: gene label, val: kcat scaling factor)
        :param orig_kcats_fname: path/file name of Excel spreadsheet with originao model kcats
        :type orig_kcats_fname: str
        :param fitted_kcats_fname: path/file name of Excel spreadsheet with scaled kcats
        :type fitted_kcats_fname: str
        """

        # load model kcats from which to extract selected transporter enzymes
        with pd.ExcelFile(orig_kcats_fname) as xlsx:
            df_orig_kcats = pd.read_excel(xlsx, sheet_name='kcats', index_col=0)
            print(f'{len(df_orig_kcats):4d} original kcat records loaded from {orig_kcats_fname}')

        gene2records = defaultdict(list)
        for record_id, row in df_orig_kcats.iterrows():
            genes = [item.strip() for item in row['info_genes'].split(',')]
            for gene in genes:
                gene2records[gene].append(record_id)
        gene2records = dict(gene2records)
        # (46) gene2records = {'JCVISYN3A_0641': ['R_5FTHFabc', 'R_COAabc', 'R_NACabc', 'R_P5Pabc', 'R_RIBFLVabc'], ...}

        record2fittedkcats = defaultdict(list)
        for gene, scale in gene2scale.items():
            for record_id in gene2records[gene]:
                orig_kcat = df_orig_kcats.at[record_id, 'kcat_per_s']
                record2fittedkcats[record_id].append(orig_kcat * scale)

        # get a unique fitted kcat value (applicable to enzyme complexes): min, max, mean or geometric mean
        # print(f'{" ":42}  min    max   mean  geo_mean')
        record2fittedkcat = {}
        for record_id, kcats in record2fittedkcats.items():
            if re.match(r'R_\w{3}t2r_', record_id) is None:
                # kcat = min(kcats)
                # kcat = max(kcats)
                # kcat = np.mean(kcats)
                kcat = np.exp(sum([np.log(kcat) for kcat in kcats])/len(kcats))
                kcat_rounded = min(1000.0, max(0.0001, round(kcat, 4)))
                record2fittedkcat[record_id] = kcat_rounded

        # create a new kcats file with updated values
        df_fitted_kcats = df_orig_kcats.copy()
        for record_id, kcat in record2fittedkcat.items():
            df_fitted_kcats.at[record_id, 'kcat_per_s'] = kcat
            df_fitted_kcats.at[record_id, 'notes'] = 'automatically fitted to proteomics'
        print(f'{len(record2fittedkcat):4d} kcats fitted')

        # export fitted kcats file
        with pd.ExcelWriter(fitted_kcats_fname) as writer:
            df_fitted_kcats.to_excel(writer, sheet_name='kcats')
            print(f'fitted kcats exported to {fitted_kcats_fname}')



"""Implementation of SbmlUnitDef class.


based on XbaUnitDef from xbanalysis package

Peter Schubert, HHU Duesseldorf, February 2023
"""

import numpy as np
import sbmlxdf

from .sbml_sbase import SbmlSBase


class SbmlUnitDef(SbmlSBase):

    def __init__(self, s_unit_def):
        super().__init__(s_unit_def)
        self.units = [SbmlUnit(sbmlxdf.extract_params(u))
                      for u in sbmlxdf.record_generator(s_unit_def['units'])]

    def is_equivalent(self, qry_units):
        if len(qry_units) != len(self.units):
            return False
        else:
            equivalent_u = np.zeros(len(qry_units))
            for i, qry_unit in enumerate(qry_units):
                for unit in self.units:
                    if unit == qry_unit:
                        equivalent_u[i] = 1
                        break
            return np.all(equivalent_u)

    def to_dict(self):
        data = super().to_dict()
        units = []
        for unit in self.units:
            units.append(', '.join([f'{key}={val}' for key, val in unit.__dict__.items()]))
        data['units'] = '; '.join(units)
        return data


class SbmlUnit:

    def __init__(self, params):
        self.kind = params.get('kind')
        self.exp = float(params.get('exp', '1.0'))
        self.scale = int(params.get('scale', '0'))
        self.mult = float(params.get('mult', '1.0'))

    def __eq__(self, other):
        if (self.kind == other.kind and
                self.exp == other.exp and
                self.scale == other.scale and
                self.mult == other.mult):
            return True
        else:
            return False

    def __str__(self):
        factor = self.mult * 10**self.scale
        factor_str = f'{factor} ' if factor != 1.0 else ''
        if self.exp != 1.0:
            unit_str = f'({factor_str}{self.kind})^{self.exp}'
        else:
            unit_str = f'{factor_str}{self.kind}'
        return unit_str



"""Implementation of UniprotProtein class.

Holds protein related information extracted from Uniprot export.

Peter Schubert, CCB, HHU Duesseldorf, January 2023
"""
import re


def get_loci(loci_str):
    """Extract locus information from Uniprot ordered locus string.

    :param loci_str: value of Uniprot 'Gene Names (ordered locus)'
    :type loci_str: str
    :return: list of loci
    :rtype: list of str
    """
    loci = []
    if type(loci_str) == str:
        loci = [re.sub(';', '', locus).strip() for locus in loci_str.split()]
    return loci


def get_location(location_str):
    """Extract compartment from Uniprot subcellular location string.

    Use first part following 'SUBCELLULR LOCATION: ', accepting only
    letter and spaces in the compartment name

    :param location_str: value of Uniprot 'Subcellular location [CC]'
    :type location_str: str
    :return: subcellular location
    :rtype: str
    """
    location = ''
    if type(location_str) == str:
        match = re.match(r'SUBCELLULAR LOCATION: ([a-zA-Z ]*)', location_str)
        if match is not None:
            location = match.group(1).strip()
    return location


def get_go_terms(go_annotation):
    """Extract Gene Ontology terms from a GO annotation element.

    Remove GO ids in square brackets
    Strip leading/trailing spaces

    :param go_annotation: value of Uniprot 'Gene Ontology (xxx)' annotation
    :type go_annotation: str
    :return: list of go terms (without go ids) - sorted
    :rtype: list of str
    """
    go_terms = []
    if type(go_annotation) is str:
        for go_term_goid in go_annotation.split(';'):
            go_term = go_term_goid.split('[')[0]
            go_terms.append(go_term.strip())
    return sorted(go_terms)


def get_refs(refs_str):
    """Extract ';' separated references from the references string.

    E.g. EC numbers for biocyc references

    :param refs_str: references, separated by ';'
    :type refs_str: str
    :return: list of references
    :rtype: list of str
    """
    refs = []
    if type(refs_str) == str:
        refs = [ref.strip() for ref in refs_str.split(';') if len(ref.strip()) > 0]
    return refs


def get_protein_name(protein_names_str):
    """Extract protein name from Uniprot protein names string.

    Keep first part of the protein names string upto any opening bracket.
    Strip leading/trailing spaces

    :param protein_names_str: value Uniprot 'Protein names'
    :type protein_names_str: str
    :return: protein name
    :rtype: str
    """
    name = ''
    if type(protein_names_str) == str:
        if ' (' not in protein_names_str:
            name = protein_names_str
        else:
            name = re.match(r'(.*?) \(', protein_names_str).group(1)
    return name


literal2float = {'one': 1.0, 'two': 2.0, 'three': 3.0, 'four': 4.0, 'five': 5.0, 'six': 6.0}


def get_cofactors(cofactor_str):
    """Extract cofactors with stoichiometry and Chebi id from Uniprot cofactors parameter.

    Example for P09831:
        - "COFACTOR: Name=[3Fe-4S] cluster; Xref=ChEBI:CHEBI:21137; Evidence={ECO:0000250};
           Note=Binds 1 [3Fe-4S] cluster. {ECO:0000250}; COFACTOR: Name=FAD; Xref=ChEBI:CHEBI:57692;
           Evidence={ECO:0000269|PubMed:4565085}; COFACTOR: Name=FMN; Xref=ChEBI:CHEBI:58210;
           Evidence={ECO:0000269|PubMed:4565085};"

    return: {'[3Fe-4S] cluster': {'stoic': 1.0, 'chebi': '21137'},
             'FAD': {'stoic': 1.0, 'chebi': '57692'},
             'FMN': {'stoic': 1.0, 'chebi': '58210'}}

    :param cofactor_str: value of Uniprot 'Cofactor'
    :type cofactor_str: str
    :return: cofactors with stoichiometry CHEBI mapping
    :rtype: dict of dict
    """
    cofactors = {}
    if type(cofactor_str) == str:
        for cf_data in cofactor_str.split('COFACTOR: Name='):
            if len(cf_data) > 0:
                name = cf_data.split(';')[0]
                m = re.search(r'Xref=ChEBI:CHEBI:(\d*)', cf_data)
                chebi = m.group(1) if m else None
                stoic = 1.0
                m = re.search(r'Note=Binds (\w*)', cf_data)
                if m:
                    stoic_str = m.group(1)
                    if stoic_str.isnumeric():
                        stoic = float(stoic_str)
                    else:
                        stoic = literal2float.get(stoic_str, 1.0)
                cofactors[name] = {'stoic': stoic, 'chebi': chebi}
    return cofactors


class UniprotProtein:

    def __init__(self, s_data):
        self.id = s_data.name
        self.organism_id = s_data['Organism (ID)']
        self.gene_name = s_data['Gene Names (primary)']
        self.loci = get_loci(s_data['Gene Names (ordered locus)'])
        self.protein_name = get_protein_name(s_data['Protein names'])
        self.ec_numbers = get_refs(s_data.get('EC number'))
        self.biocyc_ids = get_refs(s_data.get('BioCyc'))
        self.location = get_location(s_data.get(['Subcellular location [CC]']))
        self.subunit = s_data['Subunit structure']
        self.go_components = get_go_terms(s_data.get('Gene Ontology (cellular component)'))
        self.go_processes = get_go_terms(s_data.get('Gene Ontology (biological process)'))
        self.go_functions = get_go_terms(s_data.get('Gene Ontology (molecular function)'))
        self.length = s_data['Length']
        self.mass = s_data['Mass']
        self.aa_sequence = s_data['Sequence']
        self.signal_peptide = s_data.get('Signal peptide')
        self.cofactors = get_cofactors(s_data.get('Cofactor'))

    def modify_attribute(self, attribute, value):
        """modify attribute value.

        :param attribute: attribute name
        :type attribute: str
        :param value: value to be configured
        :type value: str
        """
        if attribute == 'locus':
            self.loci = [value]
        else:
            setattr(self, attribute, value)



"""Implementation Utilities to calculate molecular weights.

Peter Schubert, CCB, HHU Duesseldorf, May 2024
"""

import re
import numpy as np


# Atomic weights from NIST (standard atomic weights, considering isotopic composition)
# isotopic weights for Deuterium (D) and Tritium (T) added.
# Rest 'R' assigned zero weight
# https://physics.nist.gov/cgi-bin/Compositions/stand_alone.pl

atomic_weights = {'H': 1.007975, 'He': 4.002602, 'Li': 6.9675,  'Be': 9.0121831,  'B': 10.8135,
                  'C': 12.0106, 'N': 14.006855, 'O': 15.9994, 'F': 18.998403163, 'Ne': 20.1797,
                  'Na': 22.98976928, 'Mg': 24.3055, 'Al': 26.9815385, 'Si': 28.085, 'P': 30.973761998,
                  'S': 32.0675, 'Cl': 35.4515, 'Ar': 39.948, 'K': 39.0983, 'Ca': 40.078, 'Sc': 44.955908,
                  'Ti': 47.867, 'V': 50.9415, 'Cr': 51.9961, 'Mn': 54.938044, 'Fe': 55.845, 'Co': 58.933194,
                  'Ni': 58.6934, 'Cu': 63.546, 'Zn': 65.38, 'Ga': 69.723, 'Ge': 72.63, 'As': 74.921595,
                  'Se': 78.971, 'Br': 79.904, 'Kr': 83.798, 'Rb': 85.4678, 'Sr': 87.62, 'Y': 88.90584,
                  'Zr': 91.224, 'Nb': 92.90637, 'Mo': 95.95, 'Ru': 101.07, 'Rh': 102.9055, 'Pd': 106.42,
                  'Ag': 107.8682, 'Cd': 112.414, 'In': 114.818, 'Sn': 118.71, 'Sb': 121.76, 'Te': 127.6,
                  'I': 126.90447, 'Xe': 131.293, 'Cs': 132.90545196, 'Ba': 137.327, 'La': 138.90547,
                  'Ce': 140.116, 'Pr': 140.90766, 'Nd': 144.242, 'Sm': 150.36, 'Eu': 151.964, 'Gd': 157.25,
                  'Tb': 158.92535, 'Dy': 162.5, 'Ho': 164.93033, 'Er': 167.259, 'Tm': 168.93422, 'Yb': 173.054,
                  'Lu': 174.9668, 'Hf': 178.49, 'Ta': 180.94788, 'W': 183.84, 'Re': 186.207, 'Os': 190.23,
                  'Ir': 192.217, 'Pt': 195.084, 'Au': 196.966569, 'Hg': 200.592, 'Tl': 204.3835, 'Pb': 207.2,
                  'Bi': 208.9804, 'Th': 232.0377, 'Pa': 231.03588, 'U': 238.02891,
                  'D': 2.014102, 'T': 3.01605, 'R': 0.0}

# calculate molecular weights based on chemical formula
atom_regex_pattern = re.compile('([A-Z][a-z]*)([0-9]*)')


def get_seq_composition(seq_str):
    """Get composition of nucleotide or amino acid sequence string

    Sequence string is a string of single chars, e.g. 'MDEIIRQ...'

    :param seq_str: sequence string of single chars
    :type seq_str: str
    :return: sequence compostion
    :rtype: dict (key: char, val: count)
    """
    return {seq_char: seq_str.count(seq_char) for seq_char in sorted(set(seq_str))}


def extract_atoms(formula):
    """Iterator to return atom quantities in chemical formula.
    E.g. 'C6H12O6'
    """
    if type(formula) is str:
        for x in atom_regex_pattern.finditer(formula):
            atom = x.group(1)
            atom_stoic = float('1' if x.group(2) == '' else x.group(2))
            yield atom, atom_stoic


def calc_mw_from_formula(formula):
    """Calclulate molecular weight based on chemical formula

    using NIST atomic weights table (standard atomic weight):
        https://physics.nist.gov/cgi-bin/Compositions/stand_alone.pl

    E.g. 'C10H12N5O7P' for AMP -> 345.050 g/mol

    :param formula: chemical formula, e.g. 'H2O'
    :type formula: str
    :return: molecular weight in Da (g/mol)
    :rtype: float
    """
    composition = {atom: stoic for atom, stoic in extract_atoms(formula)}
    weight = 0.0
    for atom, stoic in composition.items():
        weight += atomic_weights.get(atom, 0.0) * stoic
    return weight


# average isotopic mass used in Expasy Compute pI/Mw tool (https://web.expasy.org/compute_pi/)
# data copied from https://web.expasy.org/findmod/findmod_masses.html#AA
# polymerized amino acids considered, e.g. L-Alanine 89.09 g/mol,
#  in protein H2O is removed per peptide bond: i.e. A = 89.09 - 18.01 = 71.08 g/mol
# Note: mean aa_weight: 110.74 g/mol (major 20 aa, H2O rmoved, weighted as per Kozlowski, 2016, pubmed: 27789699)
#  note: uknown aa could be replaced by L (most frequent)
aa_mw = {'A': 71.0788, 'C': 103.1388, 'D': 115.0886, 'E': 129.1155, 'F': 147.1766,
         'G': 57.0519, 'H': 137.1411, 'I': 113.1594, 'K': 128.1741, 'L': 113.1594,
         'M': 131.1926, 'N': 114.1038, 'P': 97.1167, 'Q': 128.1307, 'R': 156.1875,
         'S': 87.0782, 'T': 101.1051, 'V': 99.1326, 'W': 186.2132, 'Y': 163.176,
         'O': 237.3018, 'U': 150.0388,   # Pyrrolysine (Pyl, O), Selenocysteine (Sec, U)
         }
aa_freq = {'A': 8.76, 'C': 1.38, 'D': 5.49, 'E': 6.32, 'F': 3.87,
           'G': 7.03, 'H': 2.26, 'I': 5.49, 'K': 5.19, 'L': 9.68,
           'M': 2.32, 'N': 3.93, 'P': 5.02, 'Q': 3.90, 'R': 5.78,
           'S': 7.14, 'T': 5.53, 'V': 6.73, 'W': 1.25, 'Y': 2.91}
aa_avg_mw = sum([freq/100.0 * aa_mw[aa] for aa, freq in aa_freq.items()])  # avg mw used for unknown sequence ids
h2o_avg_mw = 18.01524   # average isotopic mass of one water molecule


def protein_mw_from_aa_comp(aa_dict):
    """Calculate protein molecular weight from on amino acid composition.

    Based on Expasy Compyte pI/Mw tool
    one H20 is removed from amino acid per peptide bond

    :param aa_dict: dictionary with amino acid one-letter code and stoichiometry
    :type aa_dict: dict (key: string of length 1, val: float)
    :return: molecular weight in g/mol (Da)
    :rtype: float
    """
    mw = h2o_avg_mw  # Expasy Compute pI/Mw adds one water molecule
    for aa, stoic in aa_dict.items():
        mw += stoic * aa_mw.get(aa, aa_avg_mw)
    return mw


def protein_mw_from_aa_seq(aa_seq):
    """Calculate protein molecular weight from on amino acid sequence.

    Based on Expasy Compyte pI/Mw tool
    one H20 is removed from amino acid per peptide bond
    unknown sequence identifiers get assigned a dummy cost of 100 g/mol

    :param aa_seq: sequence of amino acid one letter chars
    :type aa_seq: str
    :return: molecular weight in g/mol (Da)
    :rtype: float
    """
    mw = h2o_avg_mw  # Expasy Compute pI/Mw adds one water molecule
    for aa in sorted(set(aa_seq)):
        mw += aa_seq.count(aa) * aa_mw.get(aa, aa_avg_mw)
    return mw


# mw of individual nucleoside monophosphates with 'HO' removed due to condensation
#  based on deprotonated nucleoside monophosphates, e.g. AMP: 'C10H12N5O7P',
nt_weights = {'A': 328.047, 'C': 304.036, 'G': 344.042, 'U': 305.020}
nt_ukn_mw = np.mean(list(nt_weights.values()))   # molecular weight of unknown identifiers (should not appear)


def rna_mw_from_nt_comp(nt_dict):
    """Calculate RNA mw from nucleotide composition.

    :param nt_dict: nucleotide composition ('A', 'C', 'G', 'U')
    :type nt_dict: dict (key: one-letter char, val: float for stoic)
    :return: molecular weight in g/mol (Da)
    :rtype: float
    """
    mw = 0.0
    for nt, stoic in nt_dict.items():
        mw += stoic * nt_weights.get(nt, nt_ukn_mw)
    return mw


# mw of individual deoxy nucleoside monophosphates with 'HO' removed due to condensation
#  based on deprotonated deocxy nucleoside monophosphates, e.g. dAMP: 'C10H12N5O6P',
dnt_weights = {'A': 312.052, 'C': 288.041, 'G': 328.047, 'T': 303.041}
dnt_ukn_mw = np.mean(list(dnt_weights.values()))    # molecular weight of unknown dnt identifiers (should not appear)
dnt_complements = {'A': 'T', 'C': 'G', 'T': 'A', 'G': 'C'}


def ssdna_mw_from_dnt_comp(dnt_dict):
    """Calculate DNA mw from deoxy nucleotide composition (single strand).

    :param dnt_dict: deoxy nucleotide compositions ('A', 'C', 'G', 'T')
    :type dnt_dict: dict (key: one-letter char, val: float for stoic
    :return: molecular weight in g/mol (Da)
    :rtype: float
    """
    mw = 0.0
    for dnt, stoic in dnt_dict.items():
        mw += stoic * dnt_weights.get(dnt, dnt_ukn_mw)
    return mw


def dsdna_mw_from_dnt_comp(dnt_dict):
    """Calculate DNA mw from dexoy nucleotide composition (double strand).

    Adding deoxy nucleotides for complementary strang

    :param dnt_dict: deoxy nucleotide compositions ('A', 'C', 'G', 'T')
    :type dnt_dict: dict (key: one-letter char, val: float for stoic
    :return: molecular weight in g/mol (Da)
    :rtype: float
    """
    mw = 0.0
    for dnt, stoic in dnt_dict.items():
        mw += stoic * (dnt_weights.get(dnt, dnt_ukn_mw) + dnt_weights[dnt_complements.get(dnt, 'A')])
    return mw



"""Implementation of RbaInitAssign class.

Support initial assignement creation in SBML model, which
can be used in SBML model optimization

Peter Schubert, CCB, HHU Duesseldorf, January 2024
"""

import numpy as np
import pandas as pd
from collections import defaultdict


DEFAULT_INF = 1e9

# default function definitions used in RBA to SBML math mapping
func_defs = [
    {'id': 'michaelis_menten', 'name': 'michaelis_menten', 'notes': 'RBA function, see RBApy 1.0',
     'math': 'lambda(x, kcat, Km, kcat * x / (x + Km))'},
    {'id': 'michaelis_menten_sat', 'name': 'michaelis_menten_sat', 'notes': 'RBA function, see RBApy 1.0',
     'math': 'lambda(x, Km, x / (x + Km))'},
    {'id': 'michaelis_menten_ymin', 'name': 'michaelis_menten_ymin', 'notes': 'RBA function, see RBApy 1.0',
     'math': 'lambda(x, kcat, Km, y_min, max(y_min, kcat * x / (x + Km)))'},
    {'id': 'linear', 'name': 'linear', 'notes': 'RBA function, see RBApy 1.0',
     'math': ('lambda(x, y0, slope, x_min, x_max, y_min, y_max, ' +
              'min(y_max, max(y_min, y0 + min(x_max, max(x_min, x)) * slope)))')},
    {'id': 'indicator', 'name': 'indicator', 'notes': 'RBA function, see RBApy 1.0',
     'math': 'lambda(x, x_min, x_max, (x > x_min) && (x < x_max))'},
]


class InitialAssignments:

    def __init__(self, parameters):
        """Instantiate InitialAssignments

        :param parameters: function and aggregate definitions of RBA model
        :type parameters: RbaParameters
        """
        self.functions = parameters.functions
        self.aggregates = parameters.aggregates
        self.sref_data = {}
        self.var_bnd_data = {}

    def add_sref_ia(self, var_id, constr_id, rba_pid=None, math_var=None, math_const=None):
        """Add initial assignment data for species references (reaction substrates).

        :param var_id: variable id, e.g. reaction id
        :type var_id: str
        :param constr_id: constraint id, e.g. species sid
        :type constr_id: str
        :param rba_pid: RBA parameter function or aggregate id
        :type rba_pid: str
        :param math_var: mathematical string with variable names
        :type math_var: str (formated as per SBML math specification)
        :param math_const: mathematical string with constants only
        :type math_const: str (formated as per SBML math specification)
        """
        sref_id = f'{var_id}__{constr_id}'
        ad_dict = {'symbol': sref_id, 'var_id': var_id, 'constr_id': constr_id}

        constant_type = False
        if rba_pid:
            if rba_pid in self.functions:
                ad_dict['rba_funcs'] = [self.functions[rba_pid]]
                if ad_dict['rba_funcs'][0].type == 'constant':
                    constant_type = True
            elif rba_pid in self.aggregates:
                agg = self.aggregates[rba_pid]
                assert (agg.type == 'multiplication')
                ad_dict['rba_funcs'] = [self.functions[fid] for fid in agg.functions]

        if math_var:
            ad_dict['math_str'] = math_var
            constant_type = False
        elif math_const:
            ad_dict['math_str'] = math_const

        if constant_type is False:
            self.sref_data[sref_id] = AssignmentData(ad_dict)

    def add_var_bnd_ia(self, xba_bnd_pid, rba_pid=None, math_var=None, math_const=None):
        """Add initial assignment data for variable bound parameters (flux bounds).

        :param xba_bnd_pid: XBA parameter id, e.g. flux bound parameter id
        :type xba_bnd_pid: str
        :param rba_pid: RBA parameter function or aggregate id
        :type rba_pid: str
        :param math_var: mathematical string with variable names
        :type math_var: str (formated as per SBML math specification)
        :param math_const: mathematical string with constants only
        :type math_const: str (formated as per SBML math specification)
        """
        ad_dict = {'symbol': xba_bnd_pid}

        constant_type = False
        if rba_pid:
            if rba_pid in self.functions:
                ad_dict['rba_funcs'] = [self.functions[rba_pid]]
                if ad_dict['rba_funcs'][0].type == 'constant':
                    constant_type = True
            elif rba_pid in self.aggregates:
                agg = self.aggregates[rba_pid]
                assert (agg.type == 'multiplication')
                ad_dict['rba_funcs'] = [self.functions[fid] for fid in agg.functions]

        if math_var:
            ad_dict['math_str'] = math_var
            constant_type = False
        elif math_const:
            ad_dict['math_str'] = math_const

        if constant_type is False:
            self.var_bnd_data[xba_bnd_pid] = AssignmentData(ad_dict)

    def xba_implementation(self, xba_model):
        """Implement initial assignments in XBA model.

        Requires implementation of
        - species references ids on relevant reaction reactants that
          are subject to initial assignment
        - function definitions called within initial assignment math section
        - actual initial assignment component

        :param xba_model: reference to XBA model to get modified.
        :type xba_model: f2xba.XbaModel class
        """
        self._add_xba_sref_ids(xba_model)
        self._add_xba_func_defs(xba_model)
        self._add_xba_init_assign(xba_model)

    def _add_xba_sref_ids(self, xba_model):
        """Add species ref ids to selecrted reaction reactants.

        Required so we can address the sref ids via initial assignments

        :param xba_model: reference to XBA model to get modified.
        :type xba_model: f2xba.XbaModel class
        """
        var_srefs2id = defaultdict(dict)

        # collect sref ids on variable/reaction level
        for symbol_id, data in self.sref_data.items():
            var_srefs2id[data.var_id].update({data.constr_id: symbol_id})

        modify_attrs = []
        for var_id, srefs2id in var_srefs2id.items():
            modify_attrs.append([var_id, 'reaction', 'sref2id', srefs2id])

        df_modify_attrs = pd.DataFrame(modify_attrs, columns=['id', 'component', 'attribute', 'value'])
        df_modify_attrs.set_index('id', inplace=True)
        print(f"{len(df_modify_attrs):4d} variables (reactions) require srefs id for initial assignment")
        xba_model.modify_attributes(df_modify_attrs, 'reaction')

    @staticmethod
    def _add_xba_func_defs(xba_model):
        """Add function definitions to XBA model.

        Function definitions will be referenced by initial assignments

        :param xba_model: reference to XBA model to get modified.
        :type xba_model: f2xba.XbaModel class
        """
        for fd_dict in func_defs:
            xba_model.add_function_def(fd_dict)
        print(f"{len(func_defs):4d} function definitions added to XBA model")

    def _add_xba_init_assign(self, xba_model):
        """
        :param xba_model: reference to XBA model to get modified.
        :type xba_model: f2xba.XbaModel class
        """
        for symbol_id, data in self.sref_data.items():
            xba_model.add_initial_assignment({'symbol': symbol_id, 'math': data.expanded_math_str})
        for symbol_id, data in self.var_bnd_data.items():
            xba_model.add_initial_assignment({'symbol': symbol_id, 'math': data.expanded_math_str})


class AssignmentData:

    def __init__(self, ad_dict):
        self.symbol_id = ad_dict['symbol']
        self.var_id = ad_dict.get('var_id')
        self.constr_id = ad_dict.get('constr_id')
        self.rba_funcs = ad_dict.get('rba_funcs', [])
        self.math_str = ad_dict.get('math_str')
        self.expanded_math_str = self.get_math()

    def get_math(self):
        """Expand the function to a math string.

        :return: expanded_math_str
        :rtype: str
        """
        parts = []
        for rba_func in self.rba_funcs:
            parts.append(self.rba_function2math_str(rba_func))

        if self.math_str:
            parts.append(self.math_str)

        expanded_math_str = ' * '.join(parts)
        return expanded_math_str

    @staticmethod
    def rba_function2math_str(f):
        """Convert RBA function to an inline coded function string using function definitions.

        Not realy clean:
          Support for variable 'growth_rate' in michaelis mentent functions - wrt units

        Note: Needs to be updated, once function definitions are changed

        :param f: RBA function
        :type f: rba_model.parameters.RbaFunction.
        :return: math string using function from function definition
        :rtype: str
        """
        math_str = ''
        variable = f.variable

        if f.type == 'constant':
            constant = f.parameters['CONSTANT']
            math_str = f'({constant} dimensionless)'

        elif f.type == 'linear':
            y0 = f.parameters['LINEAR_CONSTANT']
            slope = f.parameters['LINEAR_COEF']
            x_min = f.parameters['X_MIN'] if np.isfinite(f.parameters['X_MIN']) else -DEFAULT_INF
            x_max = f.parameters['X_MAX'] if np.isfinite(f.parameters['X_MAX']) else DEFAULT_INF
            y_min = f.parameters['Y_MIN'] if np.isfinite(f.parameters['Y_MIN']) else -DEFAULT_INF
            y_max = f.parameters['Y_MAX'] if np.isfinite(f.parameters['Y_MAX']) else DEFAULT_INF
            math_str = (f'linear({variable}, {y0} dimensionless, {slope} hour, {x_min} per_h, {x_max} per_h, '
                        f'{y_min} dimensionless, {y_max} dimensionless)')

        elif f.type == 'michaelisMenten':
            ia_units = 'per_h' if variable == 'growth_rate' else 'mmol_per_gDW'
            kcat = f.parameters['kmax']
            km = f.parameters['Km']
            if 'Y_MIN' in f.parameters and np.isfinite(f.parameters['Y_MIN']):
                y_min = f.parameters['Y_MIN']
                math_str = (f'michaelis_menten_ymin({variable}, {kcat} dimensionless, {km} {ia_units}, '
                            f'{y_min} dimensionless)')
            elif kcat == 1.0:
                math_str = f'michaelis_menten_sat({variable}, {km} {ia_units})'
            else:
                math_str = f'michaelis_menten({variable}, {kcat} dimensionless, {km} {ia_units})'

        elif f.type == 'exponential':
            rate = f.parameters['RATE']
            math_str = f'exp({variable} * {rate} hour)'

        elif f.type == 'indicator':
            x_min = f.parameters['X_MIN'] if np.isfinite(f.parameters['X_MIN']) else -DEFAULT_INF
            x_max = f.parameters['X_MAX'] if np.isfinite(f.parameters['X_MAX']) else DEFAULT_INF
            math_str = f'indicator({variable}, {x_min} per_h, {x_max} per_h)'
        return math_str



"""Implementation of TdSpeciesData class.

Biochemical reactions deal with reactants, e.g. ATP
Reactants may be composed of several 'chemical' species, e.g. ATP4-, HATP3-, H2ATP2-
TD reactant has one-to-one relation with a model species
TD reactant is also linked to a single TD species with thermodynamics data

holding thermodynamic data for a TD species, like ATP4-
based on pyTFA thermo data

Reference: Book from Alberty, 2003, Thermodynamics for Biochemical Reactions

Peter Schubert, HHU Duesseldorf, Octobert 2023
"""
import numpy as np

DEBYE_HUCKEL_A = 0.51065        # √(l/mol) - Alberty, 2003, section 3.6
DEBYE_HUCKEL_B = 1.6            # √(l/mol) - Alberty, 2003, section 3.6
PH_MARGIN = 0.0                 # PH margin to be added to compartment pH to determine isomer group
MAX_ABS_CHARGE = 4              # maximum absolute electrical charge (to consider for activity reduction)
CPD_PROTON = 'cpd00067'        # seed id for protons (H) - not part of TD formulations

class TdSpeciesData:

    def __init__(self, td_sid, td_sdata, conv_factor=1.0):
        """Instantiate thermodynamic Metabolite data.

        Thermodynamics data for unique metabolites
        Several model species in different compartments can link to same td_metabolite.

        Energy units in thermo data get converted, if required, to kJ/mol
        Link between model species and TD species is via td_sid

        Note: pyTFA thermo data: ∆Gf˚ at pH 0, ionic strenght of 0
            molecular structure (i.e. formula and charges), in form of
            predominant ion at pH 7, zero ionic strength and 298 K; sterochemistry is ignored
            pKa estimated using MarbinBeans (see Jankowski et al. 2008)
            iJO1366 (iAF1260) charges and formula are based on pH 7.2

        td_sdata expected to have the keys:
          'name': species name
          'formula': chemical formula
          'charged_std': electical charge in standard conditions
          'mass_std': (g/mol) molecular weight in standard conditions
          'nH_std': number of protons in standard conditions
          'deltaGf_std': Standard Gibbs energy of formation (units as given in thermo data)
          'deltaGf_err': estimation error
          'pKa': list of pKa values
          'struct_cues': dict of cues in metabolite with stoichiometry
          'error': error information (str)

        Note: for strings we convert the type from numpy.str_ to str

        :param str td_sid: id of TD species (e.g. seed id)
        :param dict td_sdata: thermodynamics data record for metabolite extracted from TD database
        :param float conv_factor: optional, energy units conversion factor, default 1.0
        """
        self.id = td_sid
        self.name = str(td_sdata['name'])
        self.formula = str(td_sdata['formula'])
        self.charge_std = td_sdata['charge_std']
        self.mass_std = td_sdata['mass_std']
        self.nh_std = td_sdata['nH_std']
        self.pkas = np.array(sorted(td_sdata['pKa']))
        self.dfg0 = td_sdata['deltaGf_std'] * conv_factor if td_sdata['deltaGf_std'] < 1e6 else None
        self.dfg0_error = td_sdata['deltaGf_err'] * conv_factor if td_sdata['deltaGf_err'] < 1e6 else None
        self.error = str(td_sdata['error'])
        self.struct_cues = td_sdata['struct_cues']

    def get_std_transformed_gibbs_formation(self, compartment_ph, ionic_str, rt):
        """Calculate the standard transformed Gibbs energy of formation.

        based on pyTFA MetaboliteThermo.calcDGis()

        Using formula given by Goldberg and Tewari, 1991

        Thermodymamics of pseudoisomer groups at specified pH
        Equation 4.5-6 in Alberty's book:
            ∆fG'˚(I) = ∆fG1'˚(I) - RT ln(P(I))
            - P(I): binding polynomial (partitioning function) at specified ionic strength I
            - ∆fG1'˚(I): standard transformed Gibbs energy of formation of least protonated species of isomer group

        Equation 4.4-9 for protons:
            ∆fG_H+'˚(I) = ∆fG_H+˚(I=0) - (∆fG_H+˚(I=0) + RT ln(10^-ph))
            ∆fG_H+'˚ = RT ln(10) pH
            ∆fG_H+' = ∆fG_H+'˚ - RT ln(10) ph = 0

        :param float compartment_ph: compartment ph
        :param float ionic_str: compartment ionic strength in mol/l
        :param float rt: product RT in kJ/mol
        :returns: standard transformed Gibbs energy of formation, avg H atoms, avg charge
        :rtype: float, float, float
        """
        if self.id == CPD_PROTON:
            # actually not required, as protons get removed from TD calculations
            dfg0_tr = rt * np.log(10) * compartment_ph
            return dfg0_tr, self.nh_std, self.charge_std

        # protonation steps for metabolite up to max_ph
        deprot_steps = self._deprotonations_upto(compartment_ph + PH_MARGIN)

        # standard transformed Gibbs energy of formation
        dfg0_tr_lp = self._get_dfg0_tr_lp(compartment_ph, ionic_str, deprot_steps, rt)
        binding_polynomial_tr, avg_h_binding = self._get_binding_polynomial_tr(compartment_ph, ionic_str, deprot_steps)
        dfg0_tr = dfg0_tr_lp - rt * np.log(binding_polynomial_tr)

        # average charge and average H atoms
        avg_h_atoms_tr = self.nh_std - deprot_steps + avg_h_binding
        avg_charge_tr = self.charge_std - deprot_steps + avg_h_binding

        return dfg0_tr, avg_h_atoms_tr, avg_charge_tr

    def _deprotonations_upto(self, max_ph):
        """Determine deprotonation steps upto specified pH.

        Using pKa values, charge_std, nh_std.

        Initially assumed, that structure is uncharged below lowest pKa.
        - if there are no pka values up to max_ph, we assume that structure in correct
          protonation state
        - During deprotonation, structure becomes more negatively charged (1 unit per step)
        - if final charge is above or equal to standard charge, no deprotonation required
        - if deprotonation steps > 2: assume that struction is already at pH 7.0 and
          just add deprotonations above pH 7.0
        - (Note: this is not fully consistent, however the thermo data seems not to be
           consistent at all, as per Jankowski all structures should be in the predominant
           state as pH 7.0)
        - We only deprotonate as long there are still protons in the structure

        :param float max_ph: maximal pH for deprononation
        :return: number of required deprotonation steps from current structure
        :rtype: int
        """
        pkas_below_max = self.pkas[self.pkas < max_ph]

        total_steps = len(pkas_below_max)
        if total_steps == 0 or self.charge_std <= -total_steps:
            deprot_steps = 0
        else:
            deprot_steps = min(total_steps - (-self.charge_std), self.nh_std)
            if deprot_steps > 2:
                deprot_steps = min(sum(pkas_below_max > 7.0), self.nh_std)
        return deprot_steps

    def _transform_pkas_rev(self, ionic_str, max_ph, deprot_steps):
        """Transform pKa values in pysiological range to given ionic strength

        pKa values have to be adapted to ionic strength so we can determine the
        binding polynomial

        based on from pyTFA MetaboliteThermo.get_pka(), but modified.

        Using Alberty, 2003, equation 3.6-6:
            ln(K(I)) = ln(K(I=0)) + ln(10) A ∑ v_i z_i^2 √I / (1 + B√I)
            pKa(I) = pKa(I=0) - A ∑ v_i z_i^2 √I / (1 + B√I)

            - A, B are Debye-Huckel parameters
            - I is ionic strengh, z_i are charges, v_i are stoichiometric factors
            ∆rG'˚ = - RT ln(K')
                - K': apparent equilibrium constant
            pKa = -log10(K') = - ln(K')/ln(10)
            in 3.6-4:
            ln(10) pkA(I) = ln(10) pkA(I=0) - A ln(10) √I / (1 + B√I) * ∑v_i z_i^2
            pKa(I) = pkA(I=0) - A √I / (1 + B√I) * ∑v_i z_i^2

            E.g. for dissociation reaction [HATP^(3-)] = [H+] + [ATP^(4-)]
            ∑v_i z_i^2 = -1 * (-3)^2 + 1 * (+1)^2 + 1* (-4)^2 = 8
                       = -(-4+1)^2 + 1 + (-4)^2 = -(-4)^2 -2*(-4) - 1 + 1 + (-4)^2 = -2*(-4)
            charge at least protonated state: -4
            ∑v_i z_i^2 = -2 * charge = 8

        :param float ionic_str: compartment ionic strength in mol/l
        :param float max_ph: maximum pH for which to consider deprotonations
        :param int deprot_steps: deprotonation steps till maximum pH (>= 0)
        :returns: metabolite pkas in valid range in reverse order (highest to lowest)
        :rtype: list[float]
        """
        least_prot_charge = self.charge_std - deprot_steps
        extended_dh_factor = DEBYE_HUCKEL_A * np.sqrt(ionic_str) / (1.0 + DEBYE_HUCKEL_B * np.sqrt(ionic_str))
        pkas_below_max_rev = np.flip(self.pkas[self.pkas < max_ph])

        pkas_tr_rev = []
        for i, pka in enumerate(pkas_below_max_rev):
            sq_charges_factor = - 2 * (least_prot_charge + i)
            pka_tr = pka - sq_charges_factor * extended_dh_factor
            pkas_tr_rev.append(pka_tr)
        return pkas_tr_rev

    def _get_binding_polynomial_tr(self, compartment_ph, ionic_str, deprot_steps):
        """Calculate the transformed binding polynomial of a reactant.

        Requried for reactants consisting of multiple species with
        different protonation states in pysiological pH range.

        binding polynomial
        Alberty, 2003, equation 4.5-7:
            P = 1 + [H+]/K1 + [H+]^2/K1K2 +
            - K1, K2: equilibrium constants with K1 having the smallest value (highest pKa)

        average binding of hydrogen ion avg_NH
        Alberty, 2003, equations 1.3-7, -8, -9
        - avg_NH = [H+]/P dP/d[H+]                       (1.3-9)
        - P = 1 + [H+]/K1 + [H+]^2/K1K2 +                (1.3-8)
        - N = [H+] dP/d[H+] = [H+]/K1 + 2*[H+]^2/K1K2 +
        - NH = N/P = ([H+]/K1 + 2*[H+]^2/K1K2 + )/(1 + [H+]/K1 + [H+]^2/K1K2 + )   (1.3-7)

        :param float compartment_ph: compartment ph
        :param float ionic_str: compartment ionic strength in mol/l
        :param int deprot_steps: deprotonation steps till maximum pH (>= 0)
        :returns: binding polynomial, averge hydrogen ion binding
        :rtype: float, float
        """
        prod_eq_constants = 1.0
        polynomial = 1.0
        nominator = 0.0

        pkas_tr_rev = self._transform_pkas_rev(ionic_str, compartment_ph + PH_MARGIN, deprot_steps)
        proton_conc = 10.0 ** -compartment_ph
        for i, pka_tr in enumerate(pkas_tr_rev):
            prod_eq_constants *= 10.0 ** (-pka_tr)
            polynomial += proton_conc ** (i + 1) / prod_eq_constants
            nominator += (i + 1) * proton_conc ** (i + 1) / prod_eq_constants
        avg_h_binding = nominator/polynomial
        return polynomial, avg_h_binding

    def _get_dfg0_tr_lp(self, compartment_ph, ionic_str, deprot_steps, rt):
        """Calculate standard transformed Gibbs energy of formation for least protonated species.

        .. in an isomeric group, based on compartment pH and ionic strength.

        based on from pyTFA MetaboliteThermo.calcDGsp()

        1. calculate Gibbs energy change due to deprotonation
            - energy impact is increasing ∆fG˚(I=0) at each step with pKa value
            - we are using zero ionic strength pKa values to calculate the increase in Gibbs energy
              due to the addition of one H atom at that energy level of H+

        2. calculate transformation terms for H and due to ionic strength
        Alberty, 2003, equation 4.4-10:
            ∆fG'˚(I) = ∆fG˚(I=0) + RT ln(10) NH pH(I=0) - RT ln(10) (z^2 - NH) (A √I / (1 + B √I))
                     = ∆fG˚(I=0) + RT ln(10) NH pH(I) - RT ln(10) z^2 (A √I / (1 + B √I))

            Note: see Alberty 1.2-3: pH(I) = pH(I=0) + (A √I / (1 + B √I))
                - pH increases with increasing ionic strength; pH(I=0) = -log10([H+])
                ln(aH) = ln(γH [H+]); γH = 10^(-(A √I / (1 + B √I))); [H+] = 10^(-ph(I=0))
                ln(aH) = - ln(10) (pH(I=0) + (A √I / (1 + B √I))) = ln(10) pH(I)

        :param float compartment_ph: compartment ph
        :param float ionic_str: compartment ionic strength in mol/l
        :param int deprot_steps: deprotonation steps till maximum pH (>= 0)
        :param float rt: product RT in kJ/mol
        :returns: transported Gibbs energy of formation at given ionic strength
        :rtype: float
        """
        dfg0_lp = self.dfg0
        if deprot_steps > 0:
            pkas_below_max_rev = np.flip(self.pkas[self.pkas < compartment_ph + PH_MARGIN])
            for pka in pkas_below_max_rev[:deprot_steps]:
                dfg0_lp += rt * np.log(10.0) * pka

        extended_dh_factor = DEBYE_HUCKEL_A * np.sqrt(ionic_str) / (1.0 + DEBYE_HUCKEL_B * np.sqrt(ionic_str))
        least_prot_n_h = self.nh_std - deprot_steps
        least_prot_charge = self.charge_std - deprot_steps

        dfg0_lp_h = rt * np.log(10) * least_prot_n_h * compartment_ph
        ionic_str_tr = rt * np.log(10) * (least_prot_charge ** 2 - least_prot_n_h) * extended_dh_factor
        dfg0_tr_lp = dfg0_lp + dfg0_lp_h - ionic_str_tr

        return dfg0_tr_lp

    def modify_attribute(self, attribute, value):
        """Modify attribute value.

        :param str attribute: attribute name
        :param value: value to be configured
        :type value: str, float, int
        """
        if hasattr(self, attribute):
            setattr(self, attribute, value)
        else:
            print(f'unknown TD species attribute {attribute}')



"""Implementation of utilitys for TFA model.

Peter Schubert, HHU Duesseldorf, October 2023
"""
import re


atom_regex_pattern = re.compile('([A-Z][a-z]*)([0-9]*)')


def extract_atoms(formula):
    """Iterator to return atom quantities in chemical formula.

    E.g. 'C6H12O6'

    :param formula: chemical formula
    :type formula: str
    :return: atom id and stoichiometry
    :rtype: tuple (str, float)
    """
    if type(formula) is str:
        for x in atom_regex_pattern.finditer(formula):
            atom = x.group(1)
            atom_stoic = float('1' if x.group(2) == '' else x.group(2))
            yield atom, atom_stoic



"""Implementation of utility for bicocyc data parsing functions.

Peter Schubert, HHU Duesseldorf, May 2022
"""
import re
import numpy as np


def get_child_text(el_parent, child_tag):
    """retrieve child text data for a specified tag.

    There can be several elements with same tag,
    Text entries of all elements are retrieved and separated by '|'
    """
    text = ''
    if el_parent.find(child_tag) is not None:
        text = '| '.join([el_obj.text.strip() for el_obj in el_parent.findall(child_tag)])
    return strip_html_tags(text)


def to_float(string):
    try:
        return float(string)
    except ValueError:
        return np.nan


def to_int(string):
    try:
        return int(string)
    except ValueError:
        return 0


def strip_html_tags(string):
    return re.sub(r'<[^>]*>', '', string)


# get subtags of a specific component
def get_subtags(el_parent, child_tag):
    tags = []
    for el_child in el_parent.findall(child_tag):
        sub_tags = []
        for el_obj in el_child.findall('*'):
            sub_tags.append(el_obj.tag)
        tags.append('|'.join(sub_tags))
    return tags


def get_sub_obj_ids(el_parent, child_tag, obj_tag):
    obj_ids = []
    for el_child in el_parent.findall(child_tag):
        if el_child.find(obj_tag) is None:
            text = el_child.text
            if text is not None:
                obj_ids.append('(' + text.strip() + ')')
        else:
            for el_obj in el_child.findall(obj_tag):
                obj_ids.append(el_obj.get('frameid'))
    return obj_ids


def get_components(el_parent, component_type):
    """Retrieve components with stoichiometry

    :param el_parent: xml parent element from where to look for components
    :type el_parent: xml.etree.ElementTree.Element
    :param component_type: type of component to get information from
    :type  component_type: str
    :return: component ids with stoichiometry
    :rtype: dict (key: element id, value: stoichiometry as float)
    """
    components = {}
    for el_child in el_parent.findall('component'):
        ec_id = ''
        stoic = '1'
        for el_obj in el_child:
            if el_obj.tag == component_type:
                ec_id = el_obj.get('frameid')
            if el_obj.tag == 'coefficient':
                stoic = el_obj.text.strip()
        if ec_id != '':
            components[ec_id] = float(stoic)
    return components


# get resources
# type of resource (first 3 digits, compartment, coefficient)
def get_resources(el_parent, child_tag):
    resources = []
    for el_child in el_parent.findall(child_tag):
        ec_id = ''
        stoic = '1'
        compartment = ''
        for el_obj in el_child:
            if el_obj.tag in ['Protein', 'Compound', 'RNA']:
                ec_id = el_obj.get('frameid')
            if el_obj.tag == 'coefficient':
                stoic = el_obj.text.strip()
            if el_obj.tag == 'compartment':
                el_cco = el_obj.find('cco')
                compartment = el_cco.get('frameid')
        if ec_id != '':
            if compartment == '':
                resources.append(f'{ec_id}:{stoic}')
            else:
                resources.append(f'{ec_id}:{stoic}:{compartment}')
    return resources


def get_gene_products(el_parent, child_tag, product_type):
    """retrieve from XML tree a specified product type

    Note, there can be several protein or RNA products to a gene
    """
    products = []
    el_child = el_parent.find(child_tag)
    if el_child is not None:
        for el_obj in el_child:
            if el_obj.tag == product_type:
                product = el_obj.get('frameid')
                products.append(product)
    return products


def get_items(items_str, delim=';'):
    if type(items_str) != str:
        return []
    else:
        items = items_str.split(delim)
        for item in items:
            yield item.strip()


def extract_params(record):
    params = {}
    kv_pairs = record.split(',')
    for kv_pair in kv_pairs:
        if '=' in kv_pair:
            k, v = kv_pair.split('=')
            params[k.strip()] = v.strip()
    return params



"""Implementation of utility for mapping model parsing functions.

Peter Schubert, HHU Duesseldorf, December 2022
"""
import re
import os
import time
import numpy as np
import pandas as pd
from collections import defaultdict
import sbmlxdf


def valid_sbml_sid(component_id):
    """Make component id compliant to SBML SId.

    ensure that id starts with letter or `_`
    replace invalid characters by `_`

    Ref: The Systems Biology Markup Language (SBML):
    Language Specification for Level 3 Version 2 Core

    Definition of SiD
        letter ::= `a`..`z`,`A`..`Z`
        digit ::= `0`..`9`
        idChar ::= letter | digit | `_`
        SId ::= ( letter | `_` ) idChar*

    :param str component_id: component id
    :return: valid SBML SId
    :rtype: str
    """
    if re.match('[^a-zA-Z_]', component_id):
        component_id = '_' + component_id
    return re.sub('[^a-zA-Z0-9_]', '_', component_id)


def load_parameter_file(fname, sheet_names=None):
    """Load tables from configuration file.

    We can limit the sheet names to selected set of sheet names provided.
    If sheet names are not provide, all sheets of spreadsheet file will be loaded

    :param str fname: filename of XBA configuration file (.xlsx)
    :param list sheet_names: list of table to import (default: None)
    :return: imported tables
    :rtype: dict[pandas.DataFrame]
    """
    if os.path.exists(fname) is False:
        print(f'{fname} not found')
        raise FileNotFoundError

    params = {}
    with pd.ExcelFile(fname) as xlsx:
        for sheet in xlsx.sheet_names:
            if sheet_names is None or sheet in sheet_names:
                params[sheet] = pd.read_excel(xlsx, sheet_name=sheet, index_col=0)
        print(f'{len(params)} table(s) with parameters loaded from {fname} '
              f'({time.ctime(os.path.getmtime(fname))})')
    if len(params) == 0:
        print(f'None of {sheet_names} sheets have been found in the document')
        raise ValueError
    return params


def write_parameter_file(fname, params):
    """Write tables to configuration file.

    :param str fname: filename of XBA configuration file (.xlsx)
    :param params: tables to export
    :type params: dict[pandas.DataFrame]
    """
    with pd.ExcelWriter(fname) as writer:
        for sheet, df in params.items():
            df.to_excel(writer, sheet_name=sheet)
        print(f'{len(params)} table(s) with parameters written to  {fname}')


def get_srefs(srefs_str):
    """Extract composition from srefs string (component and stoichiometry).

    Species references string contains ';' separated records of composition.
    Each record contains ',' separated key=value pairs. Required keys are
    'species' and 'stoic'.

    :param str srefs_str: species references string with attibutes 'species' and 'stoic'
    :return: composition (components with stoichiometry
    :rtype: dict (key: species id, value: stoichiometry (float)
    """
    srefs = {}
    for sref_str in sbmlxdf.record_generator(srefs_str):
        params = sbmlxdf.extract_params(sref_str)
        srefs[params['species']] = float(params['stoic'])
    return srefs


def generate_srefs_str(stoichometric_str):
    """Generate species references from one side of reaction string.

    E.g. '2.0 M_h_e + M_mal__L_e' gets converted to
    {'species=M_h_e, stoic=2.0; species=M_mal__L_e, stoic=1.0'}

    :param stoichometric_str: stoichiometric string
    :type stoichometric_str: str
    :returns: species ids with stoichiometry
    :rtype: dict
    """
    d_srefs = {}
    for sref in stoichometric_str.split('+'):
        sref = sref.strip()
        parts = sref.split(' ')
        stoic = float(parts[0]) if len(parts) == 2 else '1.0'
        sid = parts[-1]
        if len(sid) > 0:
            d_srefs[sid] = stoic
    return '; '.join([f'species={sid}, stoic={stoic}' for sid, stoic in d_srefs.items()])


def parse_reaction_string(reaction_str):
    """Extract reactants/products sref strings, reversibility from reaction string.

    To support defining reactants and products with in a more readable format.
    Used, e.g. when reactants/products not defined separately in the dataframe
    e.g. 'M_fum_c + 2 M_h2o_c -> M_mal__L_c' for a reversible reaction
    {'reversible': True,
      reactants: 'species=M_fum_c, stoic=1.0; species=M_h2o_c, stoic=2.0',
      products: 'species=M_mal__L_c, stoic=1.0'}
    e.g. 'M_ac_e => ' for an irreversible reaction with no product

    :param str reaction_str: reaction string
    :returns: dict with reactions string converted to species refs string
    :rtype: dict with keys 'reversible', 'reactants', 'products'
    """
    srefs = {}
    if type(reaction_str) is str:
        if ('->' in reaction_str) or ('=>' in reaction_str):
            components = re.split(r'[=-]>', reaction_str)
        else:
            components = ['', '']
        srefs['reversible'] = ('->' in reaction_str)
        srefs['reactants'] = generate_srefs_str(components[0])
        srefs['products'] = generate_srefs_str(components[1])
    return srefs


def stoicstr2srefs(stoichometric_str):
    """Generate species references from one side of reaction string.

    E.g. '2.0 M_h_e + M_mal__L_e' transformed to
    {'M_h_e': 2.0, 'M_mal__L_e': 1.0}

    :param str stoichometric_str: stoichiometric string
    :returns: species with stoichiometry
    :rtype: dict (key: species/str, val: stoic/float)
    """
    srefs = {}
    components = [item.strip() for item in stoichometric_str.split('+')]
    for component in components:
        if ' ' in component:
            _stoic, _sid = component.split(' ')
            sid = _sid.strip()
            stoic = float(_stoic)
        else:
            sid = component
            stoic = 1.0
        srefs[sid] = stoic
    return srefs


def update_master_kcats(df_master_kcats, fname):
    """Iterative creation of kcats table.

    Master table of kcats for enzyme catalyzed reactions is getting updated
    with information from kcats table stored in fname.
    Updated master table is returned.
    Initial master table can be created using XbaModel.export_kcats(fname).

    f2xba kcats Excel Spreadsheet format:
        - sheet name: 'kcats'
        - first column is header
        - first row is the index (not required)

    f2xba kcats table format:
    index: unique record index, e.g 'R_GLUDy_iso2_REV' proposed rid, isoenzyme number, '_REV' for reverse dir.
    mandatory columns:
        - rid: reaction id, extracted from metabolic model, e.g. 'R_GLUDy'
        - dirxn: reaction direction 1 (forward) or -1 (reverse) int
        - enzyme: enzyme id, concatenation of sorted gene loci, str, e.g. 'enz_b3212_b3213'
        - kcat_per_s: kcat per second for one active site (IUPAC), float
        - notes: notes related to the entry, str or None
    optional columns:
        - info_active sites: number of active sites of enzyme (int or float)
        - info_ecns: EX numbers, comma separated, str or None, e.g. '1.4.1.13, 1.4.1.3, 1.4.1.4'
        - info_type: type of reaction, e.g. 'metabolic' or 'transporter'
        - info_genes: comma separated, e.g. 'b3212, b3213'
        - info_name: reaction name, str
        - info_reaction: reaction string in given direction, str, e.g.
            `M_akg_c + M_h_c + M_nadph_c + M_nh4_c -> M_glu__L_c + M_h2o_c + M_nadp_c`

    :param df_master_kcats: master kcats table in f2xba kcats format
    :type df_master_kcats: pandas.DataFrame
    :param str fname: relative or absolute pathname of a kcats file in f2xba kcats format
    :return: return modified master kcats table
    :rtype: pandas.DataFrame
    """
    # load data with kcats records to be updated from file.
    if os.path.exists(fname) is False:
        print(f'{fname} does not exist')
        raise FileNotFoundError
    with pd.ExcelFile(fname) as xlsx:
        df_upd_kcats = pd.read_excel(xlsx, sheet_name='kcats', index_col=0)
        print(f'{len(df_upd_kcats)} kcat records for master table updated loaded from {fname}')

    # create a mapping into the master table
    rid2isorids = defaultdict(dict)
    for iso_rid, row in df_master_kcats.iterrows():
        rid2isorids[row['rid']].update({iso_rid: [row['dirxn'], row['enzyme']]})

    not_found = {}
    updated = []
    for rkey, row in df_upd_kcats.iterrows():
        kcat = row['kcat_per_s']
        if np.isfinite(kcat) and kcat > 0.0:
            rid = row['rid']
            dirxn = row['dirxn']
            enz_id = row['enzyme']
            if rid in rid2isorids:
                idx = None
                for iso_rid, data in rid2isorids[rid].items():
                    if data[0] == dirxn:
                        if data[1] == enz_id:
                            idx = iso_rid
                            break
                if idx is None:
                    not_found[rkey] = [rid, dirxn, enz_id, kcat, row['notes']]
                else:
                    df_master_kcats.at[idx, 'kcat_per_s'] = kcat
                    if row.get('notes'):
                        df_master_kcats.at[idx, 'notes'] = row['notes']
                    if row.get('info_type'):
                        df_master_kcats.at[idx, 'info_type'] = row['info_type']
                    updated.append(idx)
            else:
                not_found[rkey] = [rid, dirxn, enz_id, kcat, row['notes']]

    print(f'{len(updated)} kcats updated, {len(df_master_kcats) - len(updated)} not updated in master; '
          f'{len(not_found)} not found in master')
    return df_master_kcats



"""Implementation of BiocycProtein class.

Peter Schubert, CCB, HHU Duesseldorf, November 2022
"""

import re
import xml.etree.ElementTree

from f2xba.utils.biocyc_utils import get_child_text, get_sub_obj_ids, get_components


class BiocycProtein:

    def __init__(self, biocyc_id):
        self.id = biocyc_id
        self.name = ''
        self.synonyms = ''
        self.gene = None
        self.enzrxns = []
        self.complexes = []
        self.protein_parts = {}
        self.rna_parts = {}
        self.compound_parts = {}
        self.gene_composition = {}

    @staticmethod
    def get_proteins(file_name):
        """Retrieve Protein data from biocyc export.

        """
        tree = xml.etree.ElementTree.parse(file_name)
        root = tree.getroot()

        data = {}
        for el in root.findall('Protein'):
            biocyc_id = el.get('ID').split(':')[1]
            bc_protein = BiocycProtein(biocyc_id)
            bc_protein.name = get_child_text(el, 'common-name')
            bc_protein.synonyms = re.sub(r'\|', ',', get_child_text(el, 'synonym'))
            genes = get_sub_obj_ids(el, 'gene', 'Gene')
            if len(genes) == 1:
                bc_protein.gene = genes[0]
            bc_protein.enzrxns = get_sub_obj_ids(el, 'catalyzes', 'Enzymatic-Reaction')
            bc_protein.complexes = get_sub_obj_ids(el, 'component-of', '*')
            bc_protein.protein_parts = get_components(el, 'Protein')
            bc_protein.rna_parts = get_components(el, 'RNA')
            bc_protein.compound_parts = get_components(el, 'Compound')

            data[biocyc_id] = bc_protein
        return data



"""Implementation of RbaEnzymes and RbaEnzyme classes.

Peter Schubert, CCB, HHU Duesseldorf, December 2022
"""

import os
import re
import pandas as pd
from xml.etree.ElementTree import parse, ElementTree, Element, SubElement, indent

from ..utils.rba_utils import get_species_refs_from_xml, get_species_refs_from_str


class RbaEnzymes:

    def __init__(self):
        self.enzymes = {}
        self.avg_enz_sat = None
        self.default_kcats = {}
        self.uptake_rcids = []
        self.saturation_sids = set()

    def import_xml(self, model_dir):

        file_name = os.path.join(model_dir, 'enzymes.xml')
        if os.path.exists(file_name) is True:
            root = parse(file_name).getroot()
            assert root.tag == 'RBAEnzymes'
            self.enzymes = RbaEnzyme.import_xml(root.find('listOfEnzymes'))
        else:
            print(f'{file_name} not found!')

    def from_df(self, m_dict):
        if 'enzymes' in m_dict:
            self.enzymes = RbaEnzyme.from_df(m_dict['enzymes'])
        else:
            print(f'enzymes not imported!')

    def create_efficiencies(self, r, active_sites, xba_model, parameters, importer_km):
        """Create enzyme fwd/rev efficiencies based on iso-reaction kcats.

        kcat is per active site. Enzymes may have multiple active sites.

        - Efficiencies have units of h-1. Reaction kcats are in s-1, therefore we multiply with 3600 s/h
          Number of active sites in the enzyme are considered.
        - For transporters taking up medium, automatically Michaelis Menten saturation functions are added
          with a default KM in mmol/l value as specified
        - For all other enzymes, assume an average enzyme saturation. Define 'avg_enz_sat' in 'general' sheet.
        Note: as import reactions all use same default Michaelis Menten Constant Km (here 1.0 mmol/l)
          external medium should be set with respect to this Km value to allow for required transporter saturation.

        E.g. Average enzyme saturation of a metabolic enzyme.
          eff = kcat * avg_enz_sat * 3600 * active_sites

        Support of RBA and TRBA.
          RBA: reversible isoreaction are not split in fwd/rev
          TRBA: reversible isoreaction are split in irreversible fwd/rev reactions

        :param r: reaction object
        :type r: f2xba.xba_model.sbml_reaction.SbmlReaction
        :param active_sites: number of active sites in the enzyme
        :type active_sites: float
        :param xba_model: xba model based on genome scale metabolic model
        :type xba_model: Class XbaModel
        :param parameters: RBA model parameters
        :type parameters: Class RbaParameters
        :param importer_km: km used for import reactions (in mmol/l)
        :type importer_km: float
        :return: enzyme efficiencies for fwd/rev direction
        :rtype: dict (keys: 'fwd', 'rev': vals: function/aggregate ids / str)
        """
        ridx = re.sub('^R_', '', r.id)

        # determine reverse catalyic rates and reverse reactants with support of both RBA and TRBA
        kcatf = r.kcatsf[0]
        kcatr = None
        rev_reactants = {}
        if r.kcatsr is not None:
            # RBA: standard case of enzyme catalyzed reversible reaction
            # TRBA: special case of enzyme catalyzed reversible reaction, if there is incomplete TD data
            kcatr = r.kcatsr[0]
            rev_reactants = r.products
        elif f'{r.id}_REV' in xba_model.reactions:
            rev_r = xba_model.reactions[f'{r.id}_REV']
            if rev_r.kcatsf is not None:
                # TRBA: normal case with reversible enzyme catalyzed reaction having complete TD data
                kcatr = rev_r.kcatsf[0]
                rev_reactants = rev_r.reactants
        kcats = {'fwd': kcatf, 'rev': kcatr}
        srefs = {'fwd': r.reactants, 'rev': rev_reactants}

        # automatically create parameters for fwd/rev enzyme efficiencies based on fwd/rev kcats
        fids = {}
        for r_dir, kcat in kcats.items():
            fid = None
            # e.g. self.uptake_rcids = {'c-e'}
            if r.compartment in self.uptake_rcids and kcat is not None:
                # for importers, the enzyme saturation is based on Michaelis Menten kinetics
                agg_fids = []
                for sid in srefs[r_dir]:
                    # e.g. self.saturation_sids = {'M_5fthf_e', 'M_ac_e', 'M_acald_e', ...} all external metabolite ids
                    if sid in self.saturation_sids:
                        agg_fid = f'saturation_{sid}'
                        if agg_fid not in parameters.functions:
                            f_params = {'type': 'michaelisMenten', 'variable': sid,
                                        'params': {'kmax': 1.0, 'Km': importer_km}}
                            parameters.add_function(agg_fid, f_params)
                        agg_fids.append(agg_fid)
                if len(agg_fids) > 0:
                    # in case of saturation terms (i.e. importer), create an RBA aggregate parameter
                    max_eff_fid = f'{ridx}_{r_dir}_eff'
                    f_params = {'type': 'constant', 'variable': 'growth_rate',
                                'params': {'CONSTANT': kcat * 3600.0 * active_sites}}
                    parameters.add_function(max_eff_fid, f_params)
                    agg_fids.append(max_eff_fid)
                    agg_id = f'{ridx}_{r_dir}_eff_agg'
                    parameters.add_aggregate(agg_id, agg_fids)
                    fid = agg_id

            if fid is None:
                # i.e. no Importer Reaction
                if kcat is not None:
                    fid = f'{ridx}_{r_dir}_eff'
                    f_params = {'type': 'constant', 'variable': 'growth_rate',
                                'params': {'CONSTANT': kcat * self.avg_enz_sat * 3600.0 * active_sites}}
                    parameters.add_function(fid, f_params)
                else:
                    fid = parameters.f_name_zero
            fids[r_dir] = fid
        return fids

    def from_xba(self, avg_enz_sat, xba_model, parameters, cid_mappings, medium, default_importer_km):
        """Configure Enzymes based on RBA sepecific parameters.

        iso-reaction kcats (s-1) are converted to enzyme efficiencies (h-1) considering and
        average enzyme saturation, and michaelis menten saturation terms for transporters

        RBA and TRBA (Thermodynamic RBA) support:
          RBA: reactions are already split in reversible isoreactions catalzyed by single enzymes.
          TRBA: reactions with TD information are split in irreversible (fwd/rev) isoreactions
          Enzyme name is based on isoreaction name prio to fwd/rev split
          i.e. RBA nad TRBA have same quantity and naming of enzymes

        Note: for reverse reaction enzyme coupling RBA and TRBA reactions have to be
          coupled differently to the enzyme.
            - RBA: C_ER_<ridx> coupled with -1 to reaction rid
            - TFBA: C_ER_<ridx> coupled with +1 to reaction rid_REV

        :param avg_enz_sat: everage enzyme saturation
        :type avg_enz_sat: float
        :param xba_model: xba model based on genome scale metabolic model
        :type xba_model: Class XbaModel
        :param parameters: RBA model parameters
        :type parameters: Class RbaParameters
        :param cid_mappings: dictionary of specific compartment id mappings
        :type cid_mappings: dict
        :param medium: Medium definition
        :type medium: class RbaMedium
        :param default_importer_km: default km used for import reactions (in mmol/l)
        :type default_importer_km: float
        """
        self.avg_enz_sat = avg_enz_sat
        self.uptake_rcids = cid_mappings['uptake_rcids']
        medium_cid = cid_mappings['medium_cid']

        # identify medium related species for saturation terms (once medium is set)
        self.saturation_sids = set()
        for mid in medium.concentrations:
            if f'{mid}_{medium_cid}' in xba_model.species:
                self.saturation_sids.add(f'{mid}_{medium_cid}')

        # get list of reactions, excluding split reverse reactions '_REV', used in TRBA
        rids = {rid for rid, r in xba_model.reactions.items() if r.kind in ['metabolic', 'transporter']}
        fwd_rids = {rid for rid in rids if re.search('_REV$', rid) is None}

        # RBA: all reactions (reversilbe/irreversible) will be in fwd_rids
        # TRBA: for enzyme determination we only iterated through the fwd reactions
        # Note: in TRBA enzyme catalzyed reactions without complete TD data are not split in fwd/rev
        for rid in fwd_rids:
            r = xba_model.reactions[rid]
            ridx = re.sub('^R_', '', rid)
            eid = f'{ridx}_enzyme'
            fids = {}
            composition = {}

            if len(r.enzymes) > 0:
                # create efficiencies for enzyme catalyzed reactions
                assert (len(r.enzymes) == 1)
                e = xba_model.enzymes[r.enzymes[0]]
                composition = e.composition
                fids = self.create_efficiencies(r, e.active_sites, xba_model, parameters, default_importer_km)
            else:
                # case of spontaneous reactions: add enzymes 1.0 to suppress warning messages in RBApy 1.0
                ub = xba_model.parameters[r.fbc_upper_bound].value
                fids['fwd'] = parameters.f_name_spontaneous if ub > 0.0 else parameters.f_name_zero
                fids['rev'] = parameters.f_name_zero
                if r.reversible is True and xba_model.parameters[r.fbc_lower_bound].value < 0.0:
                    fids['rev'] = parameters.f_name_spontaneous
                elif f'{rid}_REV' in xba_model.reactions:
                    # in case reaction has alredy been split in forward/reverse,
                    # we collect the reverse enzyme efficiency from corresponding reverse reaction
                    rev_r = xba_model.reactions[f'{rid}_REV']
                    if xba_model.parameters[rev_r.fbc_upper_bound].value > 0.0:
                        fids['rev'] = parameters.f_name_spontaneous

            if len(fids) == 2:
                rev_rid = f'{rid}_REV' if f'{rid}_REV' in xba_model.reactions else rid
                self.enzymes[eid] = RbaEnzyme(eid, rid=rid, rev_rid=rev_rid, eff_f=fids['fwd'], eff_r=fids['rev'],
                                              m_reac=composition)
        print(f'{len(self.enzymes):4d} enzymes')

    def export_xml(self, model_dir):

        file_name = os.path.join(model_dir, 'enzymes.xml')
        root = Element('RBAEnzymes')

        enzymes = SubElement(root, 'listOfEnzymes')
        for item in self.enzymes.values():
            enzymes.append(item.export_xml())

        tree = ElementTree(root)
        indent(tree)
        tree.write(file_name)

    def to_df(self):
        df = pd.DataFrame([item.to_dict() for item in self.enzymes.values()])
        df.set_index('enzyme', inplace=True)
        return {'enzymes': df}

    def validate(self, component_ids):
        valid = True
        missing = self.ref_molecules().difference(component_ids['species']) \
            .difference(component_ids['rna']) \
            .difference(component_ids['protein'])
        if len(missing) > 0:
            print('species/macromolecules used in enzyme machinery not defined:', missing)
            valid = False

        missing = self.ref_parameters().difference(component_ids['functions']) \
            .difference(component_ids['aggregates'])
        if len(missing) > 0:
            print('function/aggregates used in enzymes not defined:', missing)
            valid = False
        return valid

    def ref_molecules(self):
        refs = set()
        for e in self.enzymes.values():
            refs |= {sid for sid in e.mach_reactants}
            refs |= {sid for sid in e.mach_products}
        return refs

    def ref_parameters(self):
        refs = set()
        for e in self.enzymes.values():
            refs.add(e.forward_eff)
            refs.add(e.backward_eff)
        return refs


class RbaEnzyme:

    def __init__(self, eid, rid='', rev_rid=None, eff_f='', eff_r='', m_reac=None, m_prod=None, zero_cost=False):
        self.id = eid
        self.reaction = rid
        self.rev_reaction = rev_rid if type(rev_rid) is str else rid
        self.forward_eff = eff_f
        self.backward_eff = eff_r
        self.zero_cost = zero_cost
        self.mach_reactants = m_reac if type(m_reac) is dict else {}
        self.mach_products = m_prod if type(m_prod) is dict else {}
        self.constr_id_fwd = None
        self.constr_id_rev = None

    @staticmethod
    def import_xml(enzymes):
        data = {}
        for enzyme in enzymes.findall('enzyme'):
            eid = enzyme.attrib['id']
            rba_enzyme = RbaEnzyme(eid)
            rba_enzyme.reaction = enzyme.attrib['reaction']
            rba_enzyme.forward_eff = enzyme.attrib['forward_efficiency']
            rba_enzyme.backward_eff = enzyme.attrib['backward_efficiency']
            if enzyme.attrib.get('zeroCost', 'false').lower() == 'true':
                rba_enzyme.zero_cost = True

            machinery_composition = enzyme.find('machineryComposition')
            if machinery_composition is not None:
                rba_enzyme.mach_reactants = get_species_refs_from_xml(machinery_composition.find('listOfReactants'))
                rba_enzyme.mach_products = get_species_refs_from_xml(machinery_composition.find('listOfProducts'))
            data[eid] = rba_enzyme
        return data

    @staticmethod
    def from_df(df):
        data = {}
        for eid, row in df.iterrows():
            rba_enzyme = RbaEnzyme(eid)
            rba_enzyme.reaction = row['reaction']
            rba_enzyme.forward_eff = row['forwardEfficiency']
            rba_enzyme.backward_eff = row['backwardEfficiency']
            rba_enzyme.zero_cost = row['zeroCost']
            rba_enzyme.mach_reactants = get_species_refs_from_str(row['machineryReactants'])
            rba_enzyme.mach_products = get_species_refs_from_str(row['machineryProducts'])
            data[eid] = rba_enzyme
        return data

    def export_xml(self):
        attribs = {'id': self.id, 'reaction': self.reaction, 'forward_efficiency': self.forward_eff,
                   'backward_efficiency': self.backward_eff, 'zeroCost': str(self.zero_cost).lower()}
        enzyme = Element('enzyme', attribs)
        if len(self.mach_reactants) + len(self.mach_products) > 0:
            machinery_composition = SubElement(enzyme, 'machineryComposition')
            if len(self.mach_reactants) > 0:
                reactants = SubElement(machinery_composition, 'listOfReactants')
                for species, stoic in self.mach_reactants.items():
                    attribs = {'species': species, 'stoichiometry': str(stoic)}
                    SubElement(reactants, 'speciesReference', attribs)
            if len(self.mach_products) > 0:
                products = SubElement(machinery_composition, 'listOfProducts')
                for species, stoic in self.mach_products.items():
                    attribs = {'species': species, 'stoichiometry': str(stoic)}
                    SubElement(products, 'speciesReference', attribs)
        return enzyme

    def to_dict(self):
        mach_reactants = '; '.join([f'species={species}, stoic={stoic}'
                                    for species, stoic in self.mach_reactants.items()])
        mach_products = '; '.join([f'species={species}, stoic={stoic}'
                                   for species, stoic in self.mach_products.items()])
        return {'enzyme': self.id, 'reaction': self.reaction,
                'forwardEfficiency': self.forward_eff, 'backwardEfficiency': self.backward_eff,
                'zeroCost': self.zero_cost, 'machineryReactants': mach_reactants,
                'machineryProducts': mach_products}
